# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the intersection-over-union metric."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def evaluate(ground_truth_labels, predicted_labels, grid_size=1, name=None):
  """Computes the Intersection-Over-Union metric for the given ground truth and predicted labels.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible, and G1 to Gm are the grid dimensions.

  Args:
    ground_truth_labels: A tensor of shape `[A1, ..., An, G1, ..., Gm]`, where
      the last m axes represent a grid of ground truth attributes. Each
      attribute can either be 0 or 1.
    predicted_labels: A tensor of shape `[A1, ..., An, G1, ..., Gm]`, where the
      last m axes represent a grid of predicted attributes. Each attribute can
      either be 0 or 1.
    grid_size: The number of grid dimensions. Defaults to 1.
    name: A name for this op. Defaults to "intersection_over_union_evaluate".

  Returns:
    A tensor of shape `[A1, ..., An]` that stores the intersection-over-union
    metric of the given ground truth labels and predictions.

  Raises:
    ValueError: if the shape of `ground_truth_labels`, `predicted_labels` is
    not supported.
  """
  with tf.compat.v1.name_scope(name, "intersection_over_union_evaluate",
                               [ground_truth_labels, predicted_labels]):
    ground_truth_labels = tf.convert_to_tensor(value=ground_truth_labels)
    predicted_labels = tf.convert_to_tensor(value=predicted_labels)

    shape.compare_batch_dimensions(
        tensors=(ground_truth_labels, predicted_labels),
        tensor_names=("ground_truth_labels", "predicted_labels"),
        last_axes=-grid_size,
        broadcast_compatible=True)

    ground_truth_labels = asserts.assert_binary(ground_truth_labels)
    predicted_labels = asserts.assert_binary(predicted_labels)

    sum_ground_truth = tf.math.reduce_sum(
        input_tensor=ground_truth_labels, axis=range(-grid_size, 0))
    sum_predictions = tf.math.reduce_sum(
        input_tensor=predicted_labels, axis=range(-grid_size, 0))
    intersection = tf.math.reduce_sum(
        input_tensor=ground_truth_labels * predicted_labels,
        axis=range(-grid_size, 0))
    union = sum_ground_truth + sum_predictions - intersection

    return tf.compat.v1.where(
        tf.math.equal(union, 0), tf.ones_like(union), intersection / union)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Datasets module."""

# pylint: disable=g-import-not-at-top
from tensorflow_graphics.util.doc import _import_tfg_docs
if _import_tfg_docs():
  from tensorflow_graphics.datasets import features
  from tensorflow_graphics.datasets import modelnet40
  from tensorflow_graphics.datasets import shapenet
  from tensorflow_graphics.datasets import pix3d
  from tensorflow_graphics.util import export_api as _export_api

  # API contains submodules of tensorflow_graphics.image.
  __all__ = _export_api.get_modules()

# pylint: enable=g-import-not-at-top



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for OpenGL lookAt functions."""

from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.geometry.transformation import look_at
from tensorflow_graphics.util import test_case


class LookAtTest(test_case.TestCase):

  def test_look_at_right_handed_preset(self):
    """Tests that look_at_right_handed generates expected results."""
    camera_position = ((0.0, 0.0, 0.0), (0.1, 0.2, 0.3))
    look_at_point = ((0.0, 0.0, 1.0), (0.4, 0.5, 0.6))
    up_vector = ((0.0, 1.0, 0.0), (0.7, 0.8, 0.9))

    pred = look_at.right_handed(camera_position, look_at_point, up_vector)

    gt = (((-1.0, 0.0, 0.0, 0.0), (0.0, 1.0, 0.0, 0.0), (0.0, 0.0, -1.0, 0.0),
           (0.0, 0.0, 0.0, 1.0)),
          ((4.08248186e-01, -8.16496551e-01, 4.08248395e-01, -2.98023224e-08),
           (-7.07106888e-01, 1.19209290e-07, 7.07106769e-01, -1.41421378e-01),
           (-5.77350318e-01, -5.77350318e-01, -5.77350318e-01,
            3.46410215e-01), (0.0, 0.0, 0.0, 1.0)))
    self.assertAllClose(pred, gt)

  @parameterized.parameters(
      ((3,), (3,), (3,)),
      ((None, 3), (None, 3), (None, 3)),
      ((None, 2, 3), (None, 2, 3), (None, 2, 3)),
  )
  def test_look_at_right_handed_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(look_at.right_handed, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (2,), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (2,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3,), (1,)),
      ("Not all batch dimensions are identical", (3,), (3, 3), (3, 3)),
  )
  def test_look_at_right_handed_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(look_at.right_handed, error_msg, shapes)

  def test_look_at_right_handed_jacobian_preset(self):
    """Tests the Jacobian of look_at_right_handed."""
    camera_position_init = np.array(((0.0, 0.0, 0.0), (0.1, 0.2, 0.3)))
    look_at_init = np.array(((0.0, 0.0, 1.0), (0.4, 0.5, 0.6)))
    up_vector_init = np.array(((0.0, 1.0, 0.0), (0.7, 0.8, 0.9)))

    self.assert_jacobian_is_correct_fn(
        look_at.right_handed,
        [camera_position_init, look_at_init, up_vector_init])

  def test_look_at_right_handed_jacobian_random(self):
    """Tests the Jacobian of look_at_right_handed."""
    tensor_size = np.random.randint(1, 3)
    tensor_shape = np.random.randint(1, 5, size=(tensor_size)).tolist()
    camera_position_init = np.random.uniform(size=tensor_shape + [3])
    look_at_init = np.random.uniform(size=tensor_shape + [3])
    up_vector_init = np.random.uniform(size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(
        look_at.right_handed,
        [camera_position_init, look_at_init, up_vector_init])



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for euler-related utiliy functions."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.transformation import axis_angle
from tensorflow_graphics.geometry.transformation import euler
from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.geometry.transformation.tests import test_data as td
from tensorflow_graphics.geometry.transformation.tests import test_helpers
from tensorflow_graphics.util import test_case


class EulerTest(test_case.TestCase):

  @parameterized.parameters(
      ((3,), (1,)),
      ((None, 3), (None, 1)),
  )
  def test_from_axis_angle_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(euler.from_axis_angle, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions", (None,), (1,)),
      ("must have exactly 1 dimensions", (3,), (None,)),
  )
  def test_from_axis_angle_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(euler.from_axis_angle, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_axis_angle_jacobian_preset(self):
    """Test the Jacobian of the from_axis_angle function."""
    x_axis_init, x_angle_init = test_helpers.generate_preset_test_axis_angle()

    self.assert_jacobian_is_finite_fn(euler.from_axis_angle,
                                      [x_axis_init, x_angle_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_axis_angle_jacobian_random(self):
    """Test the Jacobian of the from_axis_angle function."""
    x_axis_init, x_angle_init = test_helpers.generate_random_test_axis_angle()

    self.assert_jacobian_is_finite_fn(euler.from_axis_angle,
                                      [x_axis_init, x_angle_init])

  def test_from_axis_angle_random(self):
    """Checks that Euler angles can be retrieved from an axis-angle."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    random_matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    random_axis, random_angle = axis_angle.from_euler(random_euler_angles)
    predicted_matrix = rotation_matrix_3d.from_axis_angle(
        random_axis, random_angle)

    self.assertAllClose(random_matrix, predicted_matrix, atol=1e-3)

  def test_from_axis_angle_preset(self):
    """Checks that Euler angles can be retrieved from axis-angle."""
    preset_euler_angles = test_helpers.generate_preset_test_euler_angles()

    random_matrix = rotation_matrix_3d.from_euler(preset_euler_angles)
    random_axis, random_angle = axis_angle.from_euler(preset_euler_angles)
    predicted_matrix = rotation_matrix_3d.from_axis_angle(
        random_axis, random_angle)

    self.assertAllClose(random_matrix, predicted_matrix, atol=1e-3)

  @parameterized.parameters(
      (td.ANGLE_90,),
      (-td.ANGLE_90,),
  )
  def test_from_axis_angle_gimbal(self, gimbal_configuration):
    """Checks that from_axis_angle works when Ry = pi/2 or -pi/2."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()
    random_euler_angles[..., 1] = gimbal_configuration

    random_matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    random_axis, random_angle = axis_angle.from_euler(random_euler_angles)
    predicted_random_angles = euler.from_axis_angle(random_axis, random_angle)
    reconstructed_random_matrices = rotation_matrix_3d.from_euler(
        predicted_random_angles)

    self.assertAllClose(reconstructed_random_matrices, random_matrix, atol=1e-3)

  @parameterized.parameters(
      ((4,),),
      ((None, 4),),
  )
  def test_from_quaternion_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(euler.from_quaternion, shape)

  @parameterized.parameters(
      ("must have exactly 4 dimensions", (None,)),)
  def test_from_quaternion_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(euler.from_quaternion, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_quaternion_jacobian_preset(self):
    """Test the Jacobian of the from_quaternion function."""
    x_init = test_helpers.generate_preset_test_quaternions()

    self.assert_jacobian_is_finite_fn(euler.from_quaternion, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_quaternion_jacobian_random(self):
    """Test the Jacobian of the from_quaternion function."""
    x_init = test_helpers.generate_random_test_quaternions()

    self.assert_jacobian_is_finite_fn(euler.from_quaternion, [x_init])

  @parameterized.parameters(
      (td.ANGLE_90,),
      (-td.ANGLE_90,),
  )
  def test_from_quaternion_gimbal(self, gimbal_configuration):
    """Checks that from_quaternion works when Ry = pi/2 or -pi/2."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()
    random_euler_angles[..., 1] = gimbal_configuration

    random_quaternion = quaternion.from_euler(random_euler_angles)
    random_matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    reconstructed_random_matrices = rotation_matrix_3d.from_quaternion(
        random_quaternion)

    self.assertAllClose(reconstructed_random_matrices, random_matrix, atol=2e-3)

  def test_from_quaternion_preset(self):
    """Checks that Euler angles can be retrieved from quaternions."""
    preset_euler_angles = test_helpers.generate_preset_test_euler_angles()

    preset_matrix = rotation_matrix_3d.from_euler(preset_euler_angles)
    preset_quaternion = quaternion.from_euler(preset_euler_angles)
    predicted_matrix = rotation_matrix_3d.from_quaternion(preset_quaternion)

    self.assertAllClose(preset_matrix, predicted_matrix, atol=2e-3)

  def test_from_quaternion_random(self):
    """Checks that Euler angles can be retrieved from quaternions."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    random_matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    random_quaternion = quaternion.from_rotation_matrix(random_matrix)
    predicted_angles = euler.from_quaternion(random_quaternion)
    predicted_matrix = rotation_matrix_3d.from_euler(predicted_angles)

    self.assertAllClose(random_matrix, predicted_matrix, atol=2e-3)

  @parameterized.parameters(
      ((3, 3),),
      ((None, 3, 3),),
  )
  def test_from_rotation_matrix_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(euler.from_rotation_matrix, shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (3,)),
      ("must have exactly 3 dimensions", (None, 3)),
      ("must have exactly 3 dimensions", (3, None)),
  )
  def test_from_rotation_matrix_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(euler.from_rotation_matrix, error_msg,
                                    shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_rotation_matrix_jacobian_preset(self):
    """Test the Jacobian of the from_rotation_matrix function."""
    x_init = test_helpers.generate_preset_test_rotation_matrices_3d()
    x = tf.convert_to_tensor(value=x_init)

    y = euler.from_rotation_matrix(x)

    self.assert_jacobian_is_finite(x, x_init, y)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_rotation_matrix_jacobian_random(self):
    """Test the Jacobian of the from_rotation_matrix function."""
    x_init = test_helpers.generate_random_test_rotation_matrix_3d()

    self.assert_jacobian_is_finite_fn(euler.from_rotation_matrix, [x_init])

  def test_from_rotation_matrix_gimbal(self):
    """Testing that Euler angles can be retrieved in Gimbal lock."""
    angles = test_helpers.generate_random_test_euler_angles()

    angles[..., 1] = np.pi / 2.
    matrix = rotation_matrix_3d.from_euler(angles)
    predicted_angles = euler.from_rotation_matrix(matrix)
    reconstructed_matrices = rotation_matrix_3d.from_euler(predicted_angles)

    self.assertAllClose(reconstructed_matrices, matrix, rtol=1e-3)

    angles[..., 1] = -np.pi / 2.
    matrix = rotation_matrix_3d.from_euler(angles)
    predicted_angles = euler.from_rotation_matrix(matrix)
    reconstructed_matrices = rotation_matrix_3d.from_euler(predicted_angles)

    self.assertAllClose(reconstructed_matrices, matrix, rtol=1e-3)

  def test_from_rotation_matrix_preset(self):
    """Tests that Euler angles can be retrieved from rotation matrices."""
    matrix = test_helpers.generate_preset_test_rotation_matrices_3d()

    predicted_angles = euler.from_rotation_matrix(matrix)
    reconstructed_matrices = rotation_matrix_3d.from_euler(predicted_angles)

    self.assertAllClose(reconstructed_matrices, matrix, rtol=1e-3)

  def test_from_rotation_matrix_random(self):
    """Tests that Euler angles can be retrieved from rotation matrices."""
    matrix = test_helpers.generate_random_test_rotation_matrix_3d()

    predicted_angles = euler.from_rotation_matrix(matrix)
    # There is not a unique mapping from rotation matrices to Euler angles. The
    # following constructs the rotation matrices from the `predicted_angles` and
    # compares them with `matrix`.
    reconstructed_matrices = rotation_matrix_3d.from_euler(predicted_angles)

    self.assertAllClose(reconstructed_matrices, matrix, rtol=1e-3)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_inverse_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(euler.inverse, shape)

  @parameterized.parameters(
      ("must have exactly 3 dimensions", (None,)),)
  def test_inverse_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(euler.inverse, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_preset(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_preset_test_euler_angles()

    self.assert_jacobian_is_correct_fn(euler.inverse, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_random(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_random_test_euler_angles()

    self.assert_jacobian_is_correct_fn(euler.inverse, [x_init])

  def test_inverse_preset(self):
    """Checks that inverse works as intended."""
    preset_euler_angles = test_helpers.generate_preset_test_euler_angles()

    prediction = euler.inverse(preset_euler_angles)
    groundtruth = -preset_euler_angles

    self.assertAllClose(prediction, groundtruth, rtol=1e-3)

  def test_inverse_random(self):
    """Checks that inverse works as intended."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    prediction = euler.inverse(random_euler_angles)
    groundtruth = -random_euler_angles

    self.assertAllClose(prediction, groundtruth, rtol=1e-3)

if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for image transformation functionalities."""

from absl.testing import parameterized
import numpy as np
import tensorflow as tf
from tensorflow_addons import image as tfa_image

from tensorflow_graphics.image import transformer
from tensorflow_graphics.util import test_case


class TransformerTest(test_case.TestCase, parameterized.TestCase):

  @parameterized.parameters(
      ((None, 1, 2, None), (None, 3, 3)),
      ((1, 2, 3, 4), (1, 3, 3)),
  )
  def test_perspective_transform_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""

    self.assert_exception_is_not_raised(transformer.perspective_transform,
                                        shape)

  @parameterized.parameters(
      ("must have a rank of 4.", (1, 1, 1), (1, 3, 3)),
      ("must have a rank of 3.", (1, 1, 1, 1), (3, 3)),
      ("Not all batch dimensions are identical.", (1, 1, 1, 1), (2, 3, 3)),
  )
  def test_perspective_transform_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""

    self.assert_exception_is_raised(transformer.perspective_transform,
                                    error_msg, shape)

  @parameterized.parameters(
      (tf.float32, "NEAREST"),
      (tf.float64, "NEAREST"),
      (tf.float32, "BILINEAR"),
      (tf.float64, "BILINEAR"),
  )
  def test_perspective_transform_half_integer_centers_preset(
      self, dtype, interpolation):
    """Tests that we can reproduce the results of tf.image.resize."""
    image = tf.constant(
        ((1.0, 2.0, 3.0), (4.0, 5.0, 6.0), (7.0, 8.0, 9.0), (10.0, 11.0, 12.0)),
        dtype=dtype)
    scale = 3
    transformation = tf.constant(
        ((1.0 / scale, 0.0, 0.0), (0.0, 1.0 / scale, 0.0), (0.0, 0.0, 1.0)),
        dtype=dtype)

    image_shape = tf.shape(image)
    image_resized_shape = image_shape * scale
    image = image[tf.newaxis, ..., tf.newaxis]
    transformation = transformation[tf.newaxis, ...]
    image_resized = tf.image.resize(
        image,
        size=image_resized_shape,
        method=tf.image.ResizeMethod.NEAREST_NEIGHBOR
        if interpolation == "NEAREST" else tf.image.ResizeMethod.BILINEAR)
    image_transformed = transformer.perspective_transform(
        image,
        transformation,
        resampling_type=transformer.ResamplingType.NEAREST
        if interpolation == "NEAREST" else transformer.ResamplingType.BILINEAR,
        border_type=transformer.BorderType.DUPLICATE,
        output_shape=image_resized_shape)

    self.assertAllClose(image_resized, image_transformed)

  @parameterized.parameters(
      (tf.float32, "NEAREST"),
      (tf.float64, "NEAREST"),
      (tf.float32, "BILINEAR"),
      (tf.float64, "BILINEAR"),
  )
  def test_perspective_transform_integer_centers_preset(self, dtype,
                                                        interpolation):
    """Tests that we can reproduce the results of tfa_image.transform."""
    image = tf.constant(
        ((1.0, 2.0, 3.0), (4.0, 5.0, 6.0), (7.0, 8.0, 9.0), (10.0, 11.0, 12.0)),
        dtype=dtype)
    scale = 3
    transformation = tf.constant(
        ((1.0 / scale, 0.0, 0.0), (0.0, 1.0 / scale, 0.0), (0.0, 0.0, 1.0)),
        dtype=dtype)

    image_shape = tf.shape(image)
    image_resized_shape = image_shape * scale
    image = image[tf.newaxis, ..., tf.newaxis]
    transformation = transformation[tf.newaxis, ...]
    image_resized = tfa_image.transform(
        tf.cast(image, tf.float32),
        tf.cast(
            tfa_image.transform_ops.matrices_to_flat_transforms(transformation),
            tf.float32),
        interpolation=interpolation,
        output_shape=image_resized_shape)
    image_transformed = transformer.perspective_transform(
        image,
        transformation,
        resampling_type=transformer.ResamplingType.NEAREST
        if interpolation == "NEAREST" else transformer.ResamplingType.BILINEAR,
        pixel_type=transformer.PixelType.INTEGER,
        output_shape=image_resized_shape)

    self.assertAllClose(image_resized, image_transformed)

  def test_perspective_transform_jacobian_random(self):
    """Tests the Jacobian of the transform function."""
    tensor_shape = np.random.randint(2, 4, size=4)
    image_init = np.random.uniform(0.0, 1.0, size=tensor_shape.tolist())
    transformation_init = np.random.uniform(
        0.0, 1.0, size=(tensor_shape[0], 3, 3))

    self.assert_jacobian_is_correct_fn(
        lambda x: transformer.perspective_transform(x, transformation_init),
        [image_init])
    self.assert_jacobian_is_correct_fn(
        lambda x: transformer.perspective_transform(image_init, x),
        [transformation_init])

  @parameterized.parameters(
      ((None, 1, 2, None), (None, 2)),
      ((1, 3, 2, 4), (1, 2)),
  )
  def test_sample_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""

    self.assert_exception_is_not_raised(transformer.sample, shape)

  @parameterized.parameters(
      ("must have a rank of 4.", (1, 1, 1), (1, 2)),
      ("must have a rank greater than 1", (1, 1, 1, 1), (2,)),
      ("Not all batch dimensions are identical.", (1, 1, 1, 1), (2, 2)),
  )
  def test_sample_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""

    self.assert_exception_is_raised(transformer.sample, error_msg, shape)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements math routines used by OpenGL."""

import tensorflow as tf

from tensorflow_graphics.geometry.transformation import look_at
from tensorflow_graphics.math.interpolation import weighted
from tensorflow_graphics.rendering.camera import perspective
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def model_to_eye(point_model_space,
                 camera_position,
                 look_at_point,
                 up_vector,
                 name=None):
  """Transforms points from model to eye coordinates.

  Note:
    In the following, A1 to An are optional batch dimensions which must be
    broadcast compatible.

  Args:
    point_model_space: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents the 3D points in model space.
    camera_position: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents the 3D position of the camera.
    look_at_point: A tensor of shape `[A1, ..., An, 3]`, with the last dimension
      storing the position where the camera is looking at.
    up_vector: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      defines the up vector of the camera.
    name: A name for this op. Defaults to 'model_to_eye'.

  Raises:
    ValueError: if the all the inputs are not of the same shape, or if any input
    of of an unsupported shape.

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, containing `point_model_space` in eye
    coordinates.
  """
  with tf.compat.v1.name_scope(
      name, "model_to_eye",
      [point_model_space, camera_position, look_at_point, up_vector]):
    point_model_space = tf.convert_to_tensor(value=point_model_space)
    camera_position = tf.convert_to_tensor(value=camera_position)
    look_at_point = tf.convert_to_tensor(value=look_at_point)
    up_vector = tf.convert_to_tensor(value=up_vector)

    shape.check_static(
        tensor=point_model_space,
        tensor_name="point_model_space",
        has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(point_model_space, camera_position),
        last_axes=-2,
        tensor_names=("point_model_space", "camera_position"),
        broadcast_compatible=True)

    model_to_eye_matrix = look_at.right_handed(camera_position, look_at_point,
                                               up_vector)
    batch_shape = tf.shape(input=point_model_space)[:-1]
    one = tf.ones(
        shape=tf.concat((batch_shape, (1,)), axis=-1),
        dtype=point_model_space.dtype)
    point_model_space = tf.concat((point_model_space, one), axis=-1)
    point_model_space = tf.expand_dims(point_model_space, axis=-1)
    res = tf.squeeze(tf.matmul(model_to_eye_matrix, point_model_space), axis=-1)
    return res[..., :-1]


def eye_to_clip(point_eye_space,
                vertical_field_of_view,
                aspect_ratio,
                near,
                far,
                name=None):
  """Transforms points from eye to clip space.

  Note:
    In the following, A1 to An are optional batch dimensions which must be
    broadcast compatible.

  Args:
    point_eye_space: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents the 3D points in eye coordinates.
    vertical_field_of_view: A tensor of shape `[A1, ..., An, 1]`, where the last
      dimension represents the vertical field of view of the frustum. Note that
      values for `vertical_field_of_view` must be in the range ]0,pi[.
    aspect_ratio: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      stores the width over height ratio of the frustum. Note that values for
      `aspect_ratio` must be non-negative.
    near: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      captures the distance between the viewer and the near clipping plane. Note
      that values for `near` must be non-negative.
    far: A tensor of shape `[A1, ..., An, 1]`, where the last dimension captures
      the distance between the viewer and the far clipping plane. Note that
      values for `far` must be non-negative.
    name: A name for this op. Defaults to 'eye_to_clip'.

  Raises:
    ValueError: If any input is of an unsupported shape.

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, containing `point_eye_space` in
    homogeneous clip coordinates.
  """
  with tf.compat.v1.name_scope(
      name, "eye_to_clip",
      [point_eye_space, vertical_field_of_view, aspect_ratio, near, far]):
    point_eye_space = tf.convert_to_tensor(value=point_eye_space)
    vertical_field_of_view = tf.convert_to_tensor(value=vertical_field_of_view)
    aspect_ratio = tf.convert_to_tensor(value=aspect_ratio)
    near = tf.convert_to_tensor(value=near)
    far = tf.convert_to_tensor(value=far)

    shape.check_static(
        tensor=point_eye_space,
        tensor_name="point_eye_space",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=vertical_field_of_view,
        tensor_name="vertical_field_of_view",
        has_dim_equals=(-1, 1))
    shape.check_static(
        tensor=aspect_ratio, tensor_name="aspect_ratio", has_dim_equals=(-1, 1))
    shape.check_static(tensor=near, tensor_name="near", has_dim_equals=(-1, 1))
    shape.check_static(tensor=far, tensor_name="far", has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(point_eye_space, vertical_field_of_view, aspect_ratio, near,
                 far),
        last_axes=-2,
        tensor_names=("point_eye_space", "vertical_field_of_view",
                      "aspect_ratio", "near", "far"),
        broadcast_compatible=True)

    perspective_matrix = perspective.right_handed(vertical_field_of_view,
                                                  aspect_ratio, near, far)
    batch_shape = tf.shape(input=point_eye_space)[:-1]
    one = tf.ones(
        shape=tf.concat((batch_shape, (1,)), axis=-1),
        dtype=point_eye_space.dtype)
    point_eye_space = tf.concat((point_eye_space, one), axis=-1)
    point_eye_space = tf.expand_dims(point_eye_space, axis=-1)

    return tf.squeeze(tf.matmul(perspective_matrix, point_eye_space), axis=-1)


def clip_to_ndc(point_clip_space, name=None):
  """Transforms points from clip to normalized device coordinates (ndc).

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    point_clip_space: A tensor of shape `[A1, ..., An, 4]`, where the last
      dimension represents points in clip space.
    name: A name for this op. Defaults to 'clip_to_ndc'.

  Raises:
    ValueError: If `point_clip_space` is not of size 4 in its last dimension.

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, containing `point_clip_space` in
    normalized device coordinates.
  """
  with tf.compat.v1.name_scope(name, "clip_to_ndc", [point_clip_space]):
    point_clip_space = tf.convert_to_tensor(value=point_clip_space)

    shape.check_static(
        tensor=point_clip_space,
        tensor_name="point_clip_space",
        has_dim_equals=(-1, 4))

    w = point_clip_space[..., -1:]
    return point_clip_space[..., :3] / w


def ndc_to_screen(point_ndc_space,
                  lower_left_corner,
                  screen_dimensions,
                  near,
                  far,
                  name=None):
  """Transforms points from normalized device coordinates to screen coordinates.

  Note:
    In the following, A1 to An are optional batch dimensions which must be
    broadcast compatible between `point_ndc_space` and the other variables.

  Args:
    point_ndc_space: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents points in normalized device coordinates.
    lower_left_corner: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension captures the position (in pixels) of the lower left corner of
      the screen.
    screen_dimensions: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension is expressed in pixels and captures the width and the height (in
      pixels) of the screen.
    near:  A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      captures the distance between the viewer and the near clipping plane. Note
      that values for `near` must be non-negative.
    far:  A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      captures the distance between the viewer and the far clipping plane. Note
      that values for `far` must be greater than those of `near`.
    name: A name for this op. Defaults to 'ndc_to_screen'.

  Raises:
    InvalidArgumentError: if any input contains data not in the specified range
      of valid values.
    ValueError: If any input is of an unsupported shape.

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, containing `point_ndc_space` in
    screen coordinates.
  """
  with tf.compat.v1.name_scope(
      name, "ndc_to_screen",
      [point_ndc_space, lower_left_corner, screen_dimensions, near, far]):
    point_ndc_space = tf.convert_to_tensor(value=point_ndc_space)
    lower_left_corner = tf.convert_to_tensor(value=lower_left_corner)
    screen_dimensions = tf.convert_to_tensor(value=screen_dimensions)
    near = tf.convert_to_tensor(value=near)
    far = tf.convert_to_tensor(value=far)

    shape.check_static(
        tensor=point_ndc_space,
        tensor_name="point_ndc_space",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=lower_left_corner,
        tensor_name="lower_left_corner",
        has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=screen_dimensions,
        tensor_name="screen_dimensions",
        has_dim_equals=(-1, 2))
    shape.check_static(tensor=near, tensor_name="near", has_dim_equals=(-1, 1))
    shape.check_static(tensor=far, tensor_name="far", has_dim_equals=(-1, 1))

    shape.compare_batch_dimensions(
        tensors=(lower_left_corner, screen_dimensions, near, far),
        last_axes=-2,
        tensor_names=("lower_left_corner", "screen_dimensions", "near", "far"),
        broadcast_compatible=False)
    shape.compare_batch_dimensions(
        tensors=(point_ndc_space, near),
        last_axes=-2,
        tensor_names=("point_ndc_space", "near"),
        broadcast_compatible=True)

    screen_dimensions = asserts.assert_all_above(
        screen_dimensions, 0.0, open_bound=True)
    near = asserts.assert_all_above(near, 0.0, open_bound=True)
    far = asserts.assert_all_above(far, near, open_bound=True)

    ndc_to_screen_factor = tf.concat(
        (screen_dimensions, far - near), axis=-1) / 2.0
    screen_center = tf.concat(
        (lower_left_corner + screen_dimensions / 2.0, (near + far) / 2.0),
        axis=-1)
    return ndc_to_screen_factor * point_ndc_space + screen_center


def model_to_screen(point_model_space,
                    model_to_eye_matrix,
                    perspective_matrix,
                    screen_dimensions,
                    lower_left_corner=(0.0, 0.0),
                    name=None):
  """Transforms points from model to screen coordinates.

  Note:
    Please refer to http://www.songho.ca/opengl/gl_transform.html for an
    in-depth review of this pipeline.

  Note:
    In the following, A1 to An are optional batch dimensions which must be
    broadcast compatible.

  Args:
    point_model_space: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents the 3D points in model space.
    model_to_eye_matrix: A tensor of shape `[A1, ..., An, 4, 4]`, where the last
      two dimension represent matrices to transform points from model to eye
      coordinates.
    perspective_matrix: A tensor of shape `[A1, ..., An, 4, 4]`, where the last
      two dimension represent matrices to transform points from eye to clip
      coordinates.
    screen_dimensions: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension is expressed in pixels and captures the width and the height (in
      pixels) of the screen.
    lower_left_corner: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension captures the position (in pixels) of the lower left corner of
      the screen.
    name: A name for this op. Defaults to 'model_to_screen'.

  Raises:
    InvalidArgumentError: if any input contains data not in the specified range
      of valid values.
    ValueError: If any input is of an unsupported shape.

  Returns:
    A tuple of two tensors, respectively of shape `[A1, ..., An, 3]` and
    `[A1, ..., An, 1]`, where the first tensor containing the projection of
    `point_model_space` in screen coordinates, and the second represents the 'w'
    component of `point_model_space` in clip space.
  """
  with tf.compat.v1.name_scope(name, "model_to_screen", [
      point_model_space, model_to_eye_matrix, perspective_matrix,
      screen_dimensions, lower_left_corner
  ]):
    point_model_space = tf.convert_to_tensor(value=point_model_space)
    model_to_eye_matrix = tf.convert_to_tensor(value=model_to_eye_matrix)
    perspective_matrix = tf.convert_to_tensor(value=perspective_matrix)

    shape.check_static(
        tensor=point_model_space,
        tensor_name="point_model_space",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=model_to_eye_matrix,
        tensor_name="model_to_eye_matrix",
        has_dim_equals=((-1, 4), (-2, 4)))
    shape.check_static(
        tensor=perspective_matrix,
        tensor_name="perspective_matrix",
        has_dim_equals=((-1, 4), (-2, 4)))
    shape.compare_batch_dimensions(
        tensors=(point_model_space, model_to_eye_matrix, perspective_matrix),
        last_axes=(-2, -3, -3),
        tensor_names=("point_model_space", "model_to_eye_matrix",
                      "perspective_matrix"),
        broadcast_compatible=True)

    batch_shape = tf.shape(input=point_model_space)[:-1]
    one = tf.ones(
        shape=tf.concat((batch_shape, (1,)), axis=-1),
        dtype=point_model_space.dtype)
    point_model_space = tf.concat((point_model_space, one), axis=-1)
    point_model_space = tf.expand_dims(point_model_space, axis=-1)

    view_projection_matrix = tf.linalg.matmul(perspective_matrix,
                                              model_to_eye_matrix)

    _, _, near, far = perspective.parameters_from_right_handed(
        perspective_matrix)

    point_clip_space = tf.squeeze(
        tf.matmul(view_projection_matrix, point_model_space), axis=-1)
    point_ndc_space = clip_to_ndc(point_clip_space)
    point_screen_space = ndc_to_screen(point_ndc_space, lower_left_corner,
                                       screen_dimensions, near, far)
    return point_screen_space, point_clip_space[..., 3:4]


def perspective_correct_barycentrics(triangle_vertices_model_space,
                                     pixel_position,
                                     model_to_eye_matrix,
                                     perspective_matrix,
                                     screen_dimensions,
                                     lower_left_corner=(0.0, 0.0),
                                     name=None):
  """Computes perspective correct barycentrics.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    triangle_vertices_model_space: A tensor of shape `[A1, ..., An, 3, 3]`,
      where the last dimension represents the vertices of a triangle in model
      space.
    pixel_position: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension stores the position (in pixels) where the interpolation is
      requested.
    model_to_eye_matrix: A tensor of shape `[A1, ..., An, 4, 4]`, where the last
      two dimension represent matrices to transform points from model to eye
      coordinates.
    perspective_matrix: A tensor of shape `[A1, ..., An, 4, 4]`, where the last
      two dimension represent matrices to transform points from eye to clip
      coordinates.
    screen_dimensions: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension is expressed in pixels and captures the width and the height (in
      pixels) of the screen.
    lower_left_corner: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension captures the position (in pixels) of the lower left corner of
      the screen.
    name: A name for this op. Defaults to 'perspective_correct_barycentrics'.

  Raises:
    InvalidArgumentError: if any input contains data not in the specified range
      of valid values.
    ValueError: If any input is of an unsupported shape.

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, containing perspective correct
    barycentric coordinates.
  """
  with tf.compat.v1.name_scope(name, "perspective_correct_barycentrics", [
      triangle_vertices_model_space, pixel_position, model_to_eye_matrix,
      perspective_matrix, screen_dimensions, lower_left_corner
  ]):
    pixel_position = tf.convert_to_tensor(value=pixel_position)
    triangle_vertices_model_space = tf.convert_to_tensor(
        value=triangle_vertices_model_space)
    shape.check_static(
        tensor=pixel_position,
        tensor_name="pixel_position",
        has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=triangle_vertices_model_space,
        tensor_name="triangle_vertices_model_space",
        has_dim_equals=((-2, 3), (-1, 3)))

    vertices_screen, vertices_w = model_to_screen(triangle_vertices_model_space,
                                                  model_to_eye_matrix,
                                                  perspective_matrix,
                                                  screen_dimensions,
                                                  lower_left_corner)
    vertices_w = tf.squeeze(vertices_w, axis=-1)
    pixel_position = tf.expand_dims(pixel_position, axis=-2)
    barycentric_coordinates, _ = weighted.get_barycentric_coordinates(
        vertices_screen[..., :2], pixel_position)
    barycentric_coordinates = tf.squeeze(barycentric_coordinates, axis=-2)
    coeffs = barycentric_coordinates / vertices_w
    return tf.linalg.normalize(coeffs, ord=1, axis=-1)[0]


def interpolate_attributes(attribute, barycentric, name=None):
  """Interpolates attributes using barycentric weights.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    attribute: A tensor of shape `[A1, ..., An, 3, B]`, where the last dimension
      stores a per-vertex `B`-dimensional attribute.
    barycentric: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      contains barycentric coordinates.
    name: A name for this op. Defaults to 'interpolate_attributes'.

  Returns:
    A tensor of shape `[A1, ..., An, B]`, containing interpolated attributes.
  """
  with tf.compat.v1.name_scope(name, "interpolate_attributes",
                               (attribute, barycentric)):
    attribute = tf.convert_to_tensor(value=attribute)
    barycentric = tf.convert_to_tensor(value=barycentric)

    shape.check_static(
        tensor=attribute, tensor_name="attribute", has_dim_equals=(-2, 3))
    shape.check_static(
        tensor=barycentric, tensor_name="barycentric", has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(attribute, barycentric),
        last_axes=(-2, -1),
        tensor_names=("attribute", "barycentric"),
        broadcast_compatible=True)
    barycentric = asserts.assert_normalized(barycentric, order=1)

    return tf.reduce_sum(
        input_tensor=tf.expand_dims(barycentric, axis=-1) * attribute, axis=-2)


def perspective_correct_interpolation(triangle_vertices_model_space,
                                      attribute,
                                      pixel_position,
                                      model_to_eye_matrix,
                                      perspective_matrix,
                                      screen_dimensions,
                                      lower_left_corner=(0.0, 0.0),
                                      name=None):
  """Returns perspective corrected interpolation of attributes over triangles.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    triangle_vertices_model_space: A tensor of shape `[A1, ..., An, 3, 3]`,
      where the last dimension represents the vertices of a triangle in model
      space.
    attribute: A tensor of shape `[A1, ..., An, 3, B]`, where the last dimension
      stores a per-vertex `B`-dimensional attribute.
    pixel_position: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension stores the position (in pixels) where the interpolation is
      requested.
    model_to_eye_matrix: A tensor of shape `[A1, ..., An, 4, 4]`, where the last
      two dimension represent matrices to transform points from model to eye
      coordinates.
    perspective_matrix: A tensor of shape `[A1, ..., An, 4, 4]`, where the last
      two dimension represent matrices to transform points from eye to clip
      coordinates.
    screen_dimensions: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension is expressed in pixels and captures the width and the height (in
      pixels) of the screen.
    lower_left_corner: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension captures the position (in pixels) of the lower left corner of
      the screen.
    name: A name for this op. Defaults to 'perspective_correct_interpolation'.

  Raises:
    tf.errors.InvalidArgumentError: if any input contains data not in the
      specified range of valid values.
    ValueError: If any input is of an unsupported shape.

  Returns:
    A tensor of shape `[A1, ..., An, B]`, containing interpolated attributes.
  """
  with tf.compat.v1.name_scope(name, "perspective_correct_interpolation", [
      triangle_vertices_model_space, attribute, pixel_position,
      model_to_eye_matrix, perspective_matrix, screen_dimensions,
      lower_left_corner
  ]):
    barycentric = perspective_correct_barycentrics(
        triangle_vertices_model_space, pixel_position, model_to_eye_matrix,
        perspective_matrix, screen_dimensions, lower_left_corner)
    return interpolate_attributes(attribute, barycentric)

# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for rotation_matrix_common."""

from absl.testing import parameterized

from tensorflow_graphics.geometry.transformation import rotation_matrix_common
from tensorflow_graphics.util import test_case


class RotationMatrixCommonTest(test_case.TestCase):

  @parameterized.parameters(
      ((2, 2),),
      ((3, 3),),
      ((4, 4),),
      ((None, 2, 2),),
      ((None, 3, 3),),
      ((None, 4, 4),),
  )
  def test_is_valid_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_not_raised(rotation_matrix_common.is_valid, shape)

  @parameterized.parameters(
      ("must have a rank greater than 1", (2,)),
      ("must have the same number of dimensions in axes", (1, 2)),
      ("must have the same number of dimensions in axes", (None, 2)),
  )
  def test_is_valid_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(rotation_matrix_common.is_valid, error_msg,
                                    shape)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Tests for ray."""

import sys

from absl import flags
from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation import ray
from tensorflow_graphics.util import test_case

FLAGS = flags.FLAGS


class RayTest(test_case.TestCase):

  def _generate_random_example(self):
    num_cameras = 4
    num_keypoints = 3
    batch_size = 2
    self.points_values = np.random.random_sample((batch_size, num_keypoints, 3))
    points_expanded_values = np.expand_dims(self.points_values, axis=-2)
    startpoints_values = np.random.random_sample(
        (batch_size, num_keypoints, num_cameras, 3))

    difference = points_expanded_values - startpoints_values
    difference_norm = np.sqrt((difference * difference).sum(axis=-1))
    direction = difference / np.expand_dims(difference_norm, axis=-1)

    self.startpoints_values = points_expanded_values - 0.5 * direction
    self.endpoints_values = points_expanded_values + 0.5 * direction
    self.weights_values = np.ones((batch_size, num_keypoints, num_cameras))

    # Wrap these with identies because some assert_* ops look at the constant
    # tensor values and mark these as unfeedable.
    self.points = tf.identity(tf.convert_to_tensor(value=self.points_values))
    self.startpoints = tf.identity(
        tf.convert_to_tensor(value=self.startpoints_values))
    self.endpoints = tf.identity(
        tf.convert_to_tensor(value=self.endpoints_values))
    self.weights = tf.identity(tf.convert_to_tensor(value=self.weights_values))

  @parameterized.parameters(
      ("Not all batch dimensions are identical.", (4, 3), (5, 3), (4,)),
      ("must have exactly 3 dimensions in axis", (4, 2), (4, 2), (4,)),
      ("must have a rank greater than 1", (3,), (3,), (None,)),
      ("must have greater than 1 dimensions in axis -2", (1, 3), (1, 3), (1,)),
      ("Not all batch dimensions are identical.", (2, 4, 3), (2, 4, 3), (2, 5)),
  )
  def test_triangulate_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(ray.triangulate, error_msg, shapes)

  @parameterized.parameters(
      ((4, 3), (4, 3), (4,)),
      ((5, 4, 3), (5, 4, 3), (5, 4)),
      ((6, 5, 4, 3), (6, 5, 4, 3), (6, 5, 4)),
  )
  def test_triangulate_exception_is_not_raised(self, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_not_raised(ray.triangulate, shapes)

  def test_triangulate_jacobian_is_correct(self):
    """Tests that Jacobian is correct."""
    self._generate_random_example()

    self.assert_jacobian_is_correct_fn(
        lambda x: ray.triangulate(x, self.endpoints, self.weights),
        [self.startpoints_values])
    self.assert_jacobian_is_correct_fn(
        lambda x: ray.triangulate(self.startpoints, x, self.weights),
        [self.endpoints_values])
    self.assert_jacobian_is_correct_fn(
        lambda x: ray.triangulate(self.startpoints, self.endpoints, x),
        [self.weights_values])

  def test_triangulate_jacobian_is_finite(self):
    """Tests that Jacobian is finite."""
    self._generate_random_example()

    self.assert_jacobian_is_finite_fn(
        lambda x: ray.triangulate(x, self.endpoints, self.weights),
        [self.startpoints_values])
    self.assert_jacobian_is_finite_fn(
        lambda x: ray.triangulate(self.startpoints, x, self.weights),
        [self.endpoints_values])
    self.assert_jacobian_is_finite_fn(
        lambda x: ray.triangulate(self.startpoints, self.endpoints, x),
        [self.weights_values])

  def test_triangulate_random(self):
    """Tests that original points are recovered by triangualtion."""
    self._generate_random_example()
    test_inputs = (self.startpoints, self.endpoints, self.weights)
    test_outputs = (self.points_values,)

    self.assert_output_is_correct(
        ray.triangulate,
        test_inputs,
        test_outputs,
        rtol=1e-05,
        atol=1e-08,
        tile=False)

  def test_negative_weights_exception_raised(self):
    """Tests that exceptions are properly raised."""
    self._generate_random_example()
    self.weights = -1.0 * tf.ones_like(self.weights, dtype=tf.float64)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      points = ray.triangulate(self.startpoints, self.endpoints, self.weights)
      self.evaluate(points)

  def test_less_that_two_nonzero_weights_exception_raised(self):
    """Tests that exceptions are properly raised."""
    self._generate_random_example()
    self.weights = tf.convert_to_tensor(
        value=np.array([[[1., 1., 0., 0.], [1., 1., 0., 0.], [1., 1., 0., 0.]],
                        [[1., 1., 0., 0.], [1., 1., 0., 0.], [1., 0., 0., 0.]]],
                       dtype=np.float64))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      points = ray.triangulate(self.startpoints, self.endpoints, self.weights)
      self.evaluate(points)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis 0", (2,), (1,), (3,), (3,)),
      ("must have a rank of 1", (2, 3), (1,), (3,), (3,)),
      ("must have exactly 1 dimensions in axis 0", (3,), (2,), (3,), (3,)),
      ("must have a rank of 1", (3,), (2, 1), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (1,), (2,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (1,), (3,), (2,)),
      ("Not all batch dimensions are identical.", (3,), (1,), (3,), (2, 3)),
  )
  def test_intersection_ray_sphere_shape_raised(self, error_msg, *shapes):
    """tests that exceptions are raised when shapes are not supported."""
    self.assert_exception_is_raised(ray.intersection_ray_sphere, error_msg,
                                    shapes)

  @parameterized.parameters(
      ((3,), (1,), (3,), (3,)),
      ((3), (1), (None, 3), (None, 3)),
  )
  def test_intersection_ray_sphere_shape_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised on supported shapes."""
    self.assert_exception_is_not_raised(ray.intersection_ray_sphere, shapes)

  def test_intersection_ray_sphere_exception_raised(self):
    """Tests that exceptions are properly raised."""
    sphere_center = np.random.uniform(size=(3,))
    point_on_ray = np.random.uniform(size=(3,))
    sample_ray = np.random.uniform(size=(3,))
    normalized_sample_ray = sample_ray / np.linalg.norm(sample_ray, axis=-1)
    positive_sphere_radius = np.random.uniform(
        sys.float_info.epsilon, 1.0, size=(1,))
    negative_sphere_radius = np.random.uniform(-1.0, 0.0, size=(1,))

    with self.subTest(name="positive_radius"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            ray.intersection_ray_sphere(sphere_center, negative_sphere_radius,
                                        normalized_sample_ray, point_on_ray))

    with self.subTest(name="normalized_ray"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            ray.intersection_ray_sphere(sphere_center, positive_sphere_radius,
                                        sample_ray, point_on_ray))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_intersection_ray_sphere_jacobian_random(self):
    """Test the Jacobian of the intersection_ray_sphere function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    sphere_center_init = np.random.uniform(0.0, 1.0, size=(3,))
    sphere_radius_init = np.random.uniform(10.0, 11.0, size=(1,))
    ray_init = np.random.uniform(size=tensor_shape + [3])
    ray_init /= np.linalg.norm(ray_init, axis=-1, keepdims=True)
    point_on_ray_init = np.random.uniform(0.0, 1.0, size=tensor_shape + [3])

    def intersection_ray_sphere_position(sphere_center, sphere_radius,
                                         input_ray, point_on_ray):
      y_p, _ = ray.intersection_ray_sphere(sphere_center, sphere_radius,
                                           input_ray, point_on_ray)
      return y_p

    def intersection_ray_sphere_normal(sphere_center, sphere_radius, input_ray,
                                       point_on_ray):
      _, y_n = ray.intersection_ray_sphere(sphere_center, sphere_radius,
                                           input_ray, point_on_ray)
      return y_n

    self.assert_jacobian_is_correct_fn(
        intersection_ray_sphere_position,
        [sphere_center_init, sphere_radius_init, ray_init, point_on_ray_init])
    self.assert_jacobian_is_correct_fn(
        intersection_ray_sphere_normal,
        [sphere_center_init, sphere_radius_init, ray_init, point_on_ray_init])

  @parameterized.parameters(
      (((0.0, 0.0, 3.0), (1.0,), (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)),
       (((0.0, 0.0, 2.0), (0.0, 0.0, 4.0)), ((0.0, 0.0, -1.0),
                                             (0.0, 0.0, 1.0)))),
      (((0.0, 0.0, 3.0), (1.0,), (0.0, 0.0, 1.0), (1.0, 0.0, 0.0)),
       (((1.0, 0.0, 3.0), (1.0, 0.0, 3.0)), ((1.0, 0.0, 0.0),
                                             (1.0, 0.0, 0.0)))),
  )
  def test_intersection_ray_sphere_preset(self, test_inputs, test_outputs):
    self.assert_output_is_correct(
        ray.intersection_ray_sphere, test_inputs, test_outputs, tile=False)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the chamfer distance loss."""

from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.nn.loss import chamfer_distance
from tensorflow_graphics.util import test_case


def _random_tensor(tensor_shape):
  return np.random.uniform(low=0.0, high=1.0, size=tensor_shape)


def _random_tensor_shape():
  tensor_size = np.random.randint(3) + 1
  return np.random.randint(1, 10, size=(tensor_size)).tolist()


def _random_point_sets():
  space_dimensions = np.random.randint(3) + 1
  batch_shape = _random_tensor_shape()
  point_set_a_size = np.random.randint(10) + 1
  point_set_b_size = np.random.randint(10) + 1

  point_set_a_init = np.random.uniform(
      low=-100.0,
      high=100.0,
      size=batch_shape + [point_set_a_size, space_dimensions])
  point_set_b_init = np.random.uniform(
      low=-100.0,
      high=100.0,
      size=batch_shape + [point_set_b_size, space_dimensions])
  return (point_set_a_init, point_set_b_init)


class ChamferDistanceTest(test_case.TestCase):

  @parameterized.parameters(
      (((0., 0), (0, 1), (1, 0), (-1, 0)),
       ((0., 0), (0, 2), (0.7, 0.4), (-0.5, -0.5)),

       # a[0] -> b[0]
       (0 + \
        # a[1] -> b[2]
        0.7**2 + 0.6**2 + \
        # a[2] -> b[2]
        0.3**2 + 0.4**2 + \
        # a[3] -> b[3]
        0.5) / 4 + \

       # b[0] -> a[0]
       (0 + \
        # b[1] -> a[1]
        1 + \
        # b[2] -> a[2]
        0.3**2 + 0.4**2 + \
        # b[3] -> a[3]
        0.5) / 4),
      (((0., 1, 4), (3, 4, 2)),
       ((2., 2, 2), (2, 3, 4), (3, 2, 2)),

       # a[0] -> b[1]
       (8 + \
        # a[1] -> b[2]
        4) / 2 + \

       # b[0] -> a[1]
       (5 + \
        # b[1] -> a[1]
        6 + \
        # b[2] -> a[1]
        4) / 3),
  )
  def test_evaluate_preset(self, point_set_a, point_set_b, expected_distance):
    tensor_shape = _random_tensor_shape()

    point_set_a = np.tile(point_set_a, tensor_shape + [1, 1])
    point_set_b = np.tile(point_set_b, tensor_shape + [1, 1])
    expected = np.tile(expected_distance, tensor_shape)

    result = chamfer_distance.evaluate(point_set_a, point_set_b)

    self.assertAllClose(expected, result)

  def test_chamfer_distance_evaluate_jacobian(self):
    """Tests the Jacobian of the Chamfer distance loss."""
    point_set_a, point_set_b = _random_point_sets()

    with self.subTest(name="jacobian_wrt_point_set_a"):
      self.assert_jacobian_is_correct_fn(
          lambda x: chamfer_distance.evaluate(x, point_set_b), [point_set_a],
          atol=1e-5)

    with self.subTest(name="jacobian_wrt_point_set_b"):
      self.assert_jacobian_is_correct_fn(
          lambda x: chamfer_distance.evaluate(point_set_a, x), [point_set_b],
          atol=1e-5)

  @parameterized.parameters(
      ("Not all batch dimensions are broadcast-compatible.", (1, 3, 5, 3),
       (2, 4, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (3, 3, 5),
       (2, 4, 5)),
      ("point_set_b must have exactly 3 dimensions in axis -1,.", (2, 4, 3),
       (2, 4, 2)),
      ("point_set_b must have exactly 2 dimensions in axis -1,.", (2, 4, 2),
       (2, 4, 3)),
  )
  def test_evaluate_shape_exception_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(chamfer_distance.evaluate, error_msg, shape)

  @parameterized.parameters(
      ((1, 5, 6, 3), (2, 5, 9, 3)),
      ((None, 2, 6, 2), (4, 2, None, 4, 2)),
      ((3, 5, 8, 7), (3, 1, 1, 7)),
  )
  def test_evaluate_shape_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(chamfer_distance.evaluate, shapes)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for exr.py."""

# pylint: disable=c-extension-no-member

import collections
import tempfile

from absl.testing import parameterized
import Imath
import numpy as np
import OpenEXR

from tensorflow_graphics.io import exr
from tensorflow_graphics.util import test_case


def _WriteMixedDatatypesExr(filename):
  header = OpenEXR.Header(64, 64)
  header['channels'] = {
      'uint': Imath.Channel(Imath.PixelType(Imath.PixelType.UINT)),
      'half': Imath.Channel(Imath.PixelType(Imath.PixelType.HALF))
  }
  channel_data = {
      'uint': np.zeros([64, 64], dtype=np.uint32).tobytes(),
      'half': np.zeros([64, 64], dtype=np.float16).tobytes()
  }
  output = OpenEXR.OutputFile(filename, header)
  output.writePixels(channel_data)
  output.close()


def _MakeTestImage(num_channels, datatype):
  if num_channels == 1:
    test_channel_names = ['L']
  else:
    test_channel_names = ['R', 'G', 'B', 'A', 'P.x', 'P.y', 'P.z']
  image = np.zeros([64, 64, num_channels], dtype=datatype)
  for i in range(num_channels):
    # Add a block of color to each channel positioned by the channel index.
    image[i:i + 10, i * 2:i * 2 + 10, i] = i / float(num_channels)
  return image, test_channel_names[:num_channels]


class ExrTest(test_case.TestCase, parameterized.TestCase):

  @parameterized.named_parameters(
      ('l', 1, np.float32, False), ('rgb', 3, np.float32, False),
      ('rgba', 4, np.float32, False), ('rgbaxyz', 7, np.float32, False),
      ('float16', 4, np.float16, False), ('unit32', 4, np.uint32, False),
      ('pass_expected_channels', 7, np.float32, True))
  def test_write_read_roundtrip(self, num_channels, datatype, pass_channels):
    expected_image, expected_channels = _MakeTestImage(num_channels, datatype)
    with tempfile.NamedTemporaryFile() as temp:
      exr.write_exr(temp.name, expected_image, expected_channels)
      image, channels = exr.read_exr(
          temp.name, expected_channels if pass_channels else None)

    self.assertEqual(expected_channels, channels)
    self.assertEqual(image.tolist(), expected_image.tolist())

  def test_reading_mixed_datatypes_fails(self):
    with tempfile.NamedTemporaryFile() as temp:
      _WriteMixedDatatypesExr(temp.name)
      with self.assertRaisesRegexp(ValueError, 'Channels have mixed datatypes'):
        _, _ = exr.read_exr(temp.name)

  def test_writing_with_array_channel_name_mismatch_fails(self):
    array_three_channels = np.zeros([64, 64, 3], dtype=np.float32)
    names_two_channels = ['A', 'B']
    with tempfile.NamedTemporaryFile() as temp:
      with self.assertRaisesRegexp(
          ValueError,
          'Number of channels in values does not match channel names'):
        exr.write_exr(temp.name, array_three_channels, names_two_channels)

  def test_writing_unsupported_numpy_type_fails(self):
    uint8_array = np.zeros([64, 64, 3], dtype=np.uint8)
    names = ['R', 'G', 'B']
    with tempfile.NamedTemporaryFile() as temp:
      with self.assertRaisesRegexp(TypeError, 'Unsupported numpy type'):
        exr.write_exr(temp.name, uint8_array, names)

  def test_reading_unknown_exr_type_fails(self):
    image, channels = _MakeTestImage(3, np.float16)
    with tempfile.NamedTemporaryFile() as temp:
      exr.write_exr(temp.name, image, channels)
      exr_file = OpenEXR.InputFile(temp.name)
      # Deliberately break the R channel header info. A mock InputFile is
      # required to override the header() method.
      header_dict = exr_file.header()
      header_dict['channels']['R'].type.v = -1  # Any bad value will do.
      make_mock_exr = collections.namedtuple('MockExr', ['header', 'channel'])
      mock_broken_exr = make_mock_exr(lambda: header_dict, exr_file.channel)
      with self.assertRaisesRegexp(RuntimeError, 'Unknown EXR channel type'):
        _ = exr.channels_to_ndarray(mock_broken_exr, ['R', 'G', 'B'])


if __name__ == '__main__':
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Tests for tensorflow_graphics.datasets.features.trimesh_feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import numpy as np
import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds
from tensorflow_graphics.datasets.features import trimesh_feature
import trimesh

_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'test_data')


class TrimeshFeatureTest(tfds.testing.FeatureExpectationsTestCase):

  def test_trimesh(self):
    obj_file_path = os.path.join(_TEST_DATA_DIR, 'cube.obj')
    obj_file = tf.io.gfile.GFile(obj_file_path)
    obj_mesh = trimesh.load(obj_file_path)
    expected_vertices = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0],
                                  [0.0, 1.0, 0.0], [0.0, 1.0, 1.0],
                                  [1.0, 0.0, 0.0], [1.0, 0.0, 1.0],
                                  [1.0, 1.0, 0.0], [1.0, 1.0, 1.0]])
    expected_faces = np.array(
        [[0, 6, 4], [0, 2, 6], [0, 3, 2], [0, 1, 3], [2, 7, 6], [2, 3, 7],
         [4, 6, 7], [4, 7, 5], [0, 4, 5], [0, 5, 1], [1, 5, 7], [1, 7, 3]],
        dtype=np.uint64)
    expected_trimesh = {'vertices': expected_vertices, 'faces': expected_faces}
    # Create a scene with two cubes.
    scene = trimesh.Scene()
    scene.add_geometry(obj_mesh)
    scene.add_geometry(obj_mesh)
    # The expected TriangleFeature for the scene.
    expected_scene_feature = {
        'vertices':
            np.tile(expected_vertices, [2, 1]).astype(np.float32),
        'faces':
            np.concatenate(
                [expected_faces, expected_faces + len(expected_vertices)],
                axis=0)
    }
    self.assertFeature(
        feature=trimesh_feature.TriangleMesh(),
        shape={
            'vertices': (None, 3),
            'faces': (None, 3)
        },
        dtype={
            'vertices': tf.float32,
            'faces': tf.uint64
        },
        tests=[
            # File path
            tfds.testing.FeatureExpectationItem(
                value=obj_file_path,
                expected=expected_trimesh,
            ),
            # File object
            tfds.testing.FeatureExpectationItem(
                value=obj_file,
                expected=expected_trimesh,
            ),
            # Trimesh
            tfds.testing.FeatureExpectationItem(
                value=obj_mesh,
                expected=expected_trimesh,
            ),
            # Scene
            tfds.testing.FeatureExpectationItem(
                value=scene,
                expected=expected_scene_feature,
            ),
            # FeaturesDict
            tfds.testing.FeatureExpectationItem(
                value=expected_scene_feature,
                expected=expected_scene_feature,
            ),
            # Invalid type
            tfds.testing.FeatureExpectationItem(
                value=np.random.rand(80, 3),
                raise_cls=ValueError,
                raise_msg='obj should be either a Trimesh or a Scene',
            ),
        ],
    )


if __name__ == '__main__':
  tfds.testing.test_main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definition of NVR+ keras model."""
import tensorflow.compat.v1 as tf
import tensorflow_graphics.projects.neural_voxel_renderer.layers as layer_utils

initializer = tf.keras.initializers.glorot_normal()
layers = tf.keras.layers


def unet_3x_with_res_in_mid(feat_in, out_filters, norm2d):
  """Helper function of a Unet with res blocks in the middle."""
  e1 = layer_utils.residual_block_2d(feat_in,
                                     nfilters=128,
                                     strides=(2, 2),
                                     normalization=norm2d)  # 16x128
  e2 = layer_utils.residual_block_2d(e1,
                                     nfilters=256,
                                     strides=(2, 2),
                                     normalization=norm2d)  # 8x256
  e3 = layer_utils.residual_block_2d(e2,
                                     nfilters=512,
                                     strides=(2, 2),
                                     normalization=norm2d)  # 4x512

  mid1 = layer_utils.residual_block_2d(e3,
                                       nfilters=512,
                                       strides=(1, 1),
                                       normalization=norm2d)  # 32x32xnf_2d
  mid2 = layer_utils.residual_block_2d(mid1,
                                       nfilters=512,
                                       strides=(1, 1),
                                       normalization=norm2d)  # 32x32xnf_2d
  mid3 = layer_utils.residual_block_2d(mid2,
                                       nfilters=512,
                                       strides=(1, 1),
                                       normalization=norm2d)  # 32x32xnf_2d
  d0 = layer_utils.upconv(mid3,
                          nfilters=256,
                          size=4,
                          strides=1)  # 8x256
  d1 = layers.concatenate([d0, e2])  # 8x512
  d2 = layers.Conv2D(256,
                     kernel_size=4,
                     strides=(1, 1),
                     padding='same',
                     kernel_initializer=initializer)(d1)  # 8x256

  d3 = layer_utils.upconv(d2,
                          nfilters=128,
                          size=4,
                          strides=1)  # 16x128
  d4 = layers.concatenate([d3, e1])  # 16x256
  d5 = layers.Conv2D(128,
                     kernel_size=4,
                     strides=(1, 1),
                     padding='same',
                     kernel_initializer=initializer)(d4)  # 8x256

  d6 = layer_utils.upconv(d5,
                          nfilters=64,
                          size=4,
                          strides=1)  # 32x64
  d7 = layers.concatenate([d6, feat_in])  # 32xN
  d8 = layers.Conv2D(out_filters,
                     kernel_size=4,
                     strides=(1, 1),
                     padding='same',
                     kernel_initializer=initializer)(d7)  # 32xout

  return d8


def neural_voxel_renderer_plus(voxels,
                               rerendering,
                               light_pos,
                               size=4,
                               norm2d='batchnorm',
                               norm3d='batchnorm'):
  """Neural Voxel Renderer + keras model."""
  with tf.name_scope('Network/'):

    voxels = layers.Input(tensor=voxels)
    rerendering = layers.Input(tensor=rerendering)
    light_pos = layers.Input(tensor=light_pos)

    nf_2d = 512

    with tf.name_scope('VoxelProcessing'):
      vol0_a = layer_utils.conv_block_3d(voxels,
                                         nfilters=16,
                                         size=size,
                                         strides=2,
                                         normalization=norm3d)  # 64x64x64x16
      vol0_b = layer_utils.conv_block_3d(vol0_a,
                                         nfilters=16,
                                         size=size,
                                         strides=1,
                                         normalization=norm3d)  # 64x64x64x16
      vol1_a = layer_utils.conv_block_3d(vol0_b,
                                         nfilters=16,
                                         size=size,
                                         strides=2,
                                         normalization=norm3d)  # 32x32x32x16
      vol1_b = layer_utils.conv_block_3d(vol1_a,
                                         nfilters=32,
                                         size=size,
                                         strides=1,
                                         normalization=norm3d)  # 32x32x32x32
      vol1_c = layer_utils.conv_block_3d(vol1_b,
                                         nfilters=32,
                                         size=size,
                                         strides=1,
                                         normalization=norm3d)  # 32x32x32x32
      shortcut = vol1_c
      vol_a1 = layer_utils.residual_block_3d(vol1_c,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a2 = layer_utils.residual_block_3d(vol_a1,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a3 = layer_utils.residual_block_3d(vol_a2,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a4 = layer_utils.residual_block_3d(vol_a3,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a5 = layer_utils.residual_block_3d(vol_a4,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      encoded_vol = layers.add([shortcut, vol_a5])
      encoded_vol = layers.Reshape([32, 32, 32*32])(encoded_vol)
      encoded_vol = layers.Conv2D(nf_2d,
                                  kernel_size=1,
                                  strides=(1, 1),
                                  padding='same',
                                  kernel_initializer=initializer)(encoded_vol)
      latent_projection = layers.LeakyReLU()(encoded_vol)  # 32x32x512

    with tf.name_scope('ProjectionProcessing'):
      shortcut = latent_projection  # 32x32xnf_2d
      e1 = layer_utils.residual_block_2d(latent_projection,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e2 = layer_utils.residual_block_2d(e1,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e3 = layer_utils.residual_block_2d(e2,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e4 = layer_utils.residual_block_2d(e3,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e5 = layer_utils.residual_block_2d(e4,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      encoded_proj = layers.add([shortcut, e5])  # 32x32xnf_2d

    with tf.name_scope('LightProcessing'):
      fc_light = layers.Dense(64, kernel_initializer=initializer)(light_pos)
      light_code = layers.Dense(64, kernel_initializer=initializer)(fc_light)
      light_code = \
        layers.Lambda(lambda v: tf.tile(v[0], [1, 32*32]))([light_code])
      light_code = layers.Reshape((32, 32, 64))(light_code)  # 32x32x64

    with tf.name_scope('Merger'):
      latent_code_final = layers.concatenate([encoded_proj, light_code])
      latent_code_final = layer_utils.conv_block_2d(latent_code_final,
                                                    nfilters=nf_2d,
                                                    size=size,
                                                    strides=1,
                                                    normalization=norm3d)
      shortcut = latent_code_final
      m1 = layer_utils.residual_block_2d(latent_code_final,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m2 = layer_utils.residual_block_2d(m1,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m3 = layer_utils.residual_block_2d(m2,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m4 = layer_utils.residual_block_2d(m3,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m5 = layer_utils.residual_block_2d(m4,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d

      latent_code_final2 = layers.add([shortcut, m5])  # 32x32xnf_2d

    with tf.name_scope('Decoder'):
      d7 = layer_utils.conv_t_block_2d(latent_code_final2,
                                       nfilters=128,
                                       size=size,
                                       strides=2,
                                       normalization=norm2d)  # 64x64x128
      d7 = layer_utils.conv_block_2d(d7,
                                     nfilters=128,
                                     size=size,
                                     strides=1,
                                     normalization=norm2d)  # 64x64x128
      d8 = layer_utils.conv_t_block_2d(d7,
                                       nfilters=64,
                                       size=size,
                                       strides=2,
                                       normalization=norm2d)  # 128x128x64
      d8 = layer_utils.conv_block_2d(d8,
                                     nfilters=64,
                                     size=size,
                                     strides=1,
                                     normalization=norm2d)  # 128x128x64
      d9 = layer_utils.conv_t_block_2d(d8,
                                       nfilters=32,
                                       size=size,
                                       strides=2,
                                       normalization=norm2d)  # 256x256x32
      d9 = layer_utils.conv_block_2d(d9,
                                     nfilters=32,
                                     size=size,
                                     strides=1,
                                     normalization=norm2d)  # 256x256x32
      rendered_image = layers.Conv2D(32,
                                     size,
                                     strides=1,
                                     padding='same',
                                     kernel_initializer=initializer,
                                     use_bias=False)(d9)  # 256x256x3

    with tf.name_scope('ImageProcessingNetwork'):
      ec1 = layer_utils.conv_block_2d(rerendering,
                                      nfilters=32,
                                      size=size,
                                      strides=1,
                                      normalization=norm2d)  # 256x
      ec2 = layer_utils.conv_block_2d(ec1,
                                      nfilters=32,
                                      size=size,
                                      strides=1,
                                      normalization=norm2d)  # 256x

    with tf.name_scope('NeuralRerenderingNetwork'):
      latent_img = layers.add([rendered_image, ec2])
      target_code = unet_3x_with_res_in_mid(latent_img, 32, norm2d=norm2d)
      out0 = layer_utils.conv_block_2d(target_code,
                                       nfilters=32,
                                       size=size,
                                       strides=1,
                                       normalization=norm2d)  # 256x
      predicted_image = layers.Conv2D(3,
                                      size,
                                      strides=1,
                                      padding='same',
                                      kernel_initializer=initializer,
                                      use_bias=False)(out0)  # 256x256x3

    return tf.keras.Model(inputs=[voxels, rerendering, light_pos],
                          outputs=[predicted_image])


def neural_voxel_renderer_plus_tf2(size=4,
                                   norm2d='batchnorm',
                                   norm3d='batchnorm'):
  """Neural Voxel Renderer + keras model for tf2."""
  with tf.name_scope('Network/'):

    voxels = layers.Input(shape=[128, 128, 128, 4])
    rerendering = layers.Input(shape=[256, 256, 3])
    light_pos = layers.Input(shape=[3])

    nf_2d = 512

    with tf.name_scope('VoxelProcessing'):
      vol0_a = layer_utils.conv_block_3d(voxels,
                                         nfilters=16,
                                         size=size,
                                         strides=2,
                                         normalization=norm3d)  # 64x64x64x16
      vol0_b = layer_utils.conv_block_3d(vol0_a,
                                         nfilters=16,
                                         size=size,
                                         strides=1,
                                         normalization=norm3d)  # 64x64x64x16
      vol1_a = layer_utils.conv_block_3d(vol0_b,
                                         nfilters=16,
                                         size=size,
                                         strides=2,
                                         normalization=norm3d)  # 32x32x32x16
      vol1_b = layer_utils.conv_block_3d(vol1_a,
                                         nfilters=32,
                                         size=size,
                                         strides=1,
                                         normalization=norm3d)  # 32x32x32x32
      vol1_c = layer_utils.conv_block_3d(vol1_b,
                                         nfilters=32,
                                         size=size,
                                         strides=1,
                                         normalization=norm3d)  # 32x32x32x32
      shortcut = vol1_c
      vol_a1 = layer_utils.residual_block_3d(vol1_c,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a2 = layer_utils.residual_block_3d(vol_a1,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a3 = layer_utils.residual_block_3d(vol_a2,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a4 = layer_utils.residual_block_3d(vol_a3,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      vol_a5 = layer_utils.residual_block_3d(vol_a4,
                                             32,
                                             strides=(1, 1, 1),
                                             normalization=norm3d)  # 32x
      encoded_vol = layers.add([shortcut, vol_a5])
      encoded_vol = layers.Reshape([32, 32, 32*32])(encoded_vol)
      encoded_vol = layers.Conv2D(nf_2d,
                                  kernel_size=1,
                                  strides=(1, 1),
                                  padding='same',
                                  kernel_initializer=initializer)(encoded_vol)
      latent_projection = layers.LeakyReLU()(encoded_vol)  # 32x32x512

    with tf.name_scope('ProjectionProcessing'):
      shortcut = latent_projection  # 32x32xnf_2d
      e1 = layer_utils.residual_block_2d(latent_projection,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e2 = layer_utils.residual_block_2d(e1,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e3 = layer_utils.residual_block_2d(e2,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e4 = layer_utils.residual_block_2d(e3,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      e5 = layer_utils.residual_block_2d(e4,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      encoded_proj = layers.add([shortcut, e5])  # 32x32xnf_2d

    with tf.name_scope('LightProcessing'):
      fc_light = layers.Dense(64, kernel_initializer=initializer)(light_pos)
      light_code = layers.Dense(64, kernel_initializer=initializer)(fc_light)
      light_code = \
        layers.Lambda(lambda v: tf.tile(v[0], [1, 32*32]))([light_code])
      light_code = layers.Reshape((32, 32, 64))(light_code)  # 32x32x64

    with tf.name_scope('Merger'):
      latent_code_final = layers.concatenate([encoded_proj, light_code])
      latent_code_final = layer_utils.conv_block_2d(latent_code_final,
                                                    nfilters=nf_2d,
                                                    size=size,
                                                    strides=1,
                                                    normalization=norm3d)
      shortcut = latent_code_final
      m1 = layer_utils.residual_block_2d(latent_code_final,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m2 = layer_utils.residual_block_2d(m1,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m3 = layer_utils.residual_block_2d(m2,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m4 = layer_utils.residual_block_2d(m3,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d
      m5 = layer_utils.residual_block_2d(m4,
                                         nfilters=nf_2d,
                                         strides=(1, 1),
                                         normalization=norm2d)  # 32x32xnf_2d

      latent_code_final2 = layers.add([shortcut, m5])  # 32x32xnf_2d

    with tf.name_scope('Decoder'):
      d7 = layer_utils.conv_t_block_2d(latent_code_final2,
                                       nfilters=128,
                                       size=size,
                                       strides=2,
                                       normalization=norm2d)  # 64x64x128
      d7 = layer_utils.conv_block_2d(d7,
                                     nfilters=128,
                                     size=size,
                                     strides=1,
                                     normalization=norm2d)  # 64x64x128
      d8 = layer_utils.conv_t_block_2d(d7,
                                       nfilters=64,
                                       size=size,
                                       strides=2,
                                       normalization=norm2d)  # 128x128x64
      d8 = layer_utils.conv_block_2d(d8,
                                     nfilters=64,
                                     size=size,
                                     strides=1,
                                     normalization=norm2d)  # 128x128x64
      d9 = layer_utils.conv_t_block_2d(d8,
                                       nfilters=32,
                                       size=size,
                                       strides=2,
                                       normalization=norm2d)  # 256x256x32
      d9 = layer_utils.conv_block_2d(d9,
                                     nfilters=32,
                                     size=size,
                                     strides=1,
                                     normalization=norm2d)  # 256x256x32
      rendered_image = layers.Conv2D(32,
                                     size,
                                     strides=1,
                                     padding='same',
                                     kernel_initializer=initializer,
                                     use_bias=False)(d9)  # 256x256x3

    with tf.name_scope('ImageProcessingNetwork'):
      ec1 = layer_utils.conv_block_2d(rerendering,
                                      nfilters=32,
                                      size=size,
                                      strides=1,
                                      normalization=norm2d)  # 256x
      ec2 = layer_utils.conv_block_2d(ec1,
                                      nfilters=32,
                                      size=size,
                                      strides=1,
                                      normalization=norm2d)  # 256x

    with tf.name_scope('NeuralRerenderingNetwork'):
      latent_img = layers.add([rendered_image, ec2])
      target_code = unet_3x_with_res_in_mid(latent_img, 32, norm2d=norm2d)
      out0 = layer_utils.conv_block_2d(target_code,
                                       nfilters=32,
                                       size=size,
                                       strides=1,
                                       normalization=norm2d)  # 256x
      predicted_image = layers.Conv2D(3,
                                      size,
                                      strides=1,
                                      padding='same',
                                      kernel_initializer=initializer,
                                      use_bias=False)(out0)  # 256x256x3

    return tf.keras.Model(inputs=[voxels, rerendering, light_pos],
                          outputs=[predicted_image])



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Tests for tensorflow_graphics.datasets.features.voxel_feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds

from tensorflow_graphics.datasets.features import voxel_feature

_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'test_data')


class VoxelGridFeatureTest(tfds.testing.FeatureExpectationsTestCase):
  """Test Cases for VoxelGrid FeatureConnector."""

  def test_voxel(self):
    """Tests voxel I/O and encoding/decoding to DatasetFeature."""

    mat_file_path = os.path.join(_TEST_DATA_DIR, 'cube.mat')
    expected_voxel = np.zeros((16, 16, 16), dtype=np.float32)
    expected_voxel[4:12, 4:12, 4:12] = 1.

    mat_dict = {'path': mat_file_path, 'key': 'voxels'}
    raising_inputs = {'path': mat_file_path, 'foo': 'voxels'}
    wrong_key = {'path': mat_file_path, 'key': 'foo'}
    wrong_path = {'path': '/somewhere/wrong', 'key': 'voxels'}
    wrong_dim = np.ones((1, 1, 1, 1))
    self.assertFeature(
        feature=voxel_feature.VoxelGrid((16, 16, 16)),
        shape=(16, 16, 16),
        dtype=tf.float32,
        tests=[
            # mat file
            tfds.testing.FeatureExpectationItem(
                value=mat_dict,
                expected=expected_voxel,
            ),
            # Voxel Grid
            tfds.testing.FeatureExpectationItem(
                value=expected_voxel,
                expected=expected_voxel,
            ),
            tfds.testing.FeatureExpectationItem(
                value=raising_inputs,
                raise_cls=ValueError,
                raise_msg='Missing keys in provided dictionary!',
            ),
            tfds.testing.FeatureExpectationItem(
                value=wrong_key,
                raise_cls=ValueError,
                raise_msg='Key `foo` not found in .mat file',
            ),
            tfds.testing.FeatureExpectationItem(
                value=wrong_path,
                raise_cls=FileNotFoundError,
                raise_msg='File `/somewhere/wrong` does not exist.',
            ),
            tfds.testing.FeatureExpectationItem(
                value=wrong_dim,
                raise_cls=ValueError,
                raise_msg='Only 3D Voxel Grids are supported.',
            ),
        ],
    )


if __name__ == '__main__':
  tfds.testing.test_main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Camera feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
from tensorflow_datasets import features

from tensorflow_graphics.datasets.features import pose_feature


class Camera(features.FeaturesDict):
  """`FeatureConnector` for camera calibration (extrinsic and intrinsic).

  During `_generate_examples`, the feature connector accepts as input:

    * `parameter_dict:` A dictionary containing the extrinsic and instrinsic
    parameters of the camera as:
      - 'pose': Dictionary containing
          * Either 3x3 rotation matrix and translation vector:
            {
              'R': A `float32` tensor with shape `[3, 3]` denoting the
                   3D rotation matrix.
              't': A `float32` tensor with shape `[3,]` denoting the
                   translation vector.
            }
      OR
         * look_at, position and up-vector:
            {
              'look_at': float32 vector of shape (3,).
              'position': float32 vector of shape (3,).
              'up': float32 vector of shape (3,).
            }
      - 'f': focal length of the camera in pixel (either single float32 value
      or tuple of float32 as (f_x, f_y).
      - 'optical_center': Optical center of the camera
      in pixel coordinates as tuple (c_x, c_y)
      Optional parameters:
      - 'skew': float32 denoting the skew of the camera axes.
      - 'aspect_ratio': float32 denoting the aspect_ratio,
      if single fixed focal length is provided.


  Output:
    A dictionary containing:

    * 'pose': A `tensorflow_graphics.datasets.features.Pose` FeatureConnector
    representing the 3D pose of the camera.
    * 'intrinsics': A `float32` tensor with shape `[3,3]` denoting the intrinsic
    matrix.

  Example:
    Default values for skew (s) and aspect_ratio(a) are 0 and 1, respectively.

    Full calibration matrix:
      K = [[ f_x,   s, c_x ],
           [   0, f_y, c_y ],
           [   0,   0,  1  ]]

    With same focal length:
      K = [[ f,  s, c_x ],
           [ 0, af, c_y ],
           [ 0,  0,  1  ]]
  """

  def __init__(self):
    super(Camera, self).__init__({
        'pose': pose_feature.Pose(),
        'intrinsics': features.Tensor(shape=(3, 3), dtype=tf.float32),
    })

  def encode_example(self, example_dict):
    """Convert the given parameters into a dict convertible to tf example."""
    REQUIRED_KEYS = ['pose', 'f', 'optical_center']  # pylint: disable=invalid-name
    if not all(key in example_dict for key in REQUIRED_KEYS):
      raise ValueError(f'Missing keys in provided dictionary! '
                       f'Expected {REQUIRED_KEYS}, '
                       f'but {example_dict.keys()} were given.')

    if not isinstance(example_dict['pose'], dict):
      raise ValueError('Pose needs to be a dictionary containing either '
                       'rotation and translation or look at, '
                       'up vector and position.')
    features_dict = {}
    pose_dict = example_dict['pose']
    if all(key in pose_dict for key in ['R', 't']):
      features_dict['pose'] = {
          'R': pose_dict['R'],
          't': pose_dict['t']
      }
    elif all(key in pose_dict for key in ['look_at', 'position', 'up']):
      rotation = self._create_rotation_from_look_at(pose_dict['look_at'],
                                                    pose_dict['position'],
                                                    pose_dict['up'])
      translation = (-rotation) @ pose_dict['position']

      features_dict['pose'] = {
          'R': rotation,
          't': translation
      }
    else:
      raise ValueError('Wrong keys for pose feature provided!')

    aspect_ratio = 1
    skew = 0
    if 'aspect_ratio' in example_dict.keys():
      if not isinstance(example_dict['f'], float):
        raise ValueError('If aspect ratio is provided, '
                         'f needs to be a single float.')
      aspect_ratio = example_dict['aspect_ratio']

    if 'skew' in example_dict.keys():
      skew = example_dict['skew']

    features_dict['intrinsics'] = self._create_calibration_matrix(
        example_dict['f'],
        example_dict['optical_center'],
        aspect_ratio,
        skew
    )

    return super(Camera, self).encode_example(features_dict)

  def _create_rotation_from_look_at(self, look_at, position, up):
    """Creates rotation matrix according to OpenGL gluLookAt convention.

    Args:
      look_at: A float32 3D vector of look_at direction.
      position: A float32 3D vector of camera position.
      up: A float32 3D up direction vector.

    Returns:
      A 3x3 float32 rotation matrix.

    (https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/gluLookAt.xml)
    """
    dir_vec = look_at - position
    dir_vec /= np.linalg.norm(dir_vec)
    side_vec = np.cross(dir_vec, up)
    side_vec /= np.linalg.norm(side_vec)
    up_vec = np.cross(side_vec, dir_vec)
    matrix = np.array([side_vec, up_vec, -dir_vec])

    return matrix

  def _create_calibration_matrix(self, f, optical_center, aspect_ratio=1,
                                 skew=0):
    """Constructs the 3x3 calibration matrix K.

    Args:
      f: Focal length of the camera. Either single float.32 value or tuple of
        float32 when different focal lengths for each axis are provided (fx, fy)
      optical_center: Tuple (c_x, c_y) containing the optical center
        of the camera in pixel coordinates.
      aspect_ratio: Optional parameter, if fixed focal length for both
        dimensions is used. Defaults to 1.
      skew: Optional parameter denoting the skew between the camera axes.

    Returns:
      float32 Tensor of shape [3,3] containing the upper triangular
      calibration matrix K.
    """
    if not isinstance(optical_center, tuple):
      raise ValueError('Optical center of camera needs '
                       'to be a tuple of (c_x, c_y).')

    if isinstance(f, tuple):
      f_x, f_y = f
    else:
      f_x = f
      f_y = aspect_ratio * f

    return np.asarray([[f_x, skew, optical_center[0]],
                       [0, f_y, optical_center[1]],
                       [0, 0, 1]
                       ], dtype=np.float32)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the fscore metric."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.nn.metric import precision as precision_module
from tensorflow_graphics.nn.metric import recall as recall_module
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def evaluate(ground_truth,
             prediction,
             precision_function=precision_module.evaluate,
             recall_function=recall_module.evaluate,
             name=None):
  """Computes the fscore metric for the given ground truth and predicted labels.

  The fscore is calculated as 2 * (precision * recall) / (precision + recall)
  where the precision and recall are evaluated by the given function parameters.
  The precision and recall functions default to their definition for boolean
  labels (see https://en.wikipedia.org/wiki/Precision_and_recall for more
  details).

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Args:
    ground_truth: A tensor of shape `[A1, ..., An, N]`, where the last axis
      represents the ground truth values.
    prediction: A tensor of shape `[A1, ..., An, N]`, where the last axis
      represents the predicted values.
    precision_function: The function to use for evaluating the precision.
      Defaults to the precision evaluation for binary ground-truth and
      predictions.
    recall_function: The function to use for evaluating the recall. Defaults to
      the recall evaluation for binary ground-truth and prediction.
    name: A name for this op. Defaults to "fscore_evaluate".

  Returns:
    A tensor of shape `[A1, ..., An]` that stores the fscore metric for the
    given ground truth labels and predictions.

  Raises:
    ValueError: if the shape of `ground_truth`, `prediction` is
    not supported.
  """
  with tf.compat.v1.name_scope(name, "fscore_evaluate",
                               [ground_truth, prediction]):
    ground_truth = tf.convert_to_tensor(value=ground_truth)
    prediction = tf.convert_to_tensor(value=prediction)

    shape.compare_batch_dimensions(
        tensors=(ground_truth, prediction),
        tensor_names=("ground_truth", "prediction"),
        last_axes=-1,
        broadcast_compatible=True)

    recall = recall_function(ground_truth, prediction)
    precision = precision_function(ground_truth, prediction)

    return safe_ops.safe_signed_div(2 * precision * recall, precision + recall)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for 2d rotation matrix."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.geometry.transformation import rotation_matrix_2d
from tensorflow_graphics.geometry.transformation.tests import test_data as td
from tensorflow_graphics.geometry.transformation.tests import test_helpers
from tensorflow_graphics.util import test_case


class RotationMatrix2dTest(test_case.TestCase):

  @parameterized.parameters(
      ((1,)),
      ((None, 1),),
  )
  def test_from_euler_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_2d.from_euler, shapes)

  @parameterized.parameters(
      ("must have exactly 1 dimensions in axis -1", (None,)),)
  def test_from_euler_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(rotation_matrix_2d.from_euler, error_msg,
                                    shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_jacobian_preset(self):
    """Test the Jacobian of the from_euler function."""
    x_init = test_helpers.generate_preset_test_euler_angles(dimensions=1)

    self.assert_jacobian_is_correct_fn(rotation_matrix_2d.from_euler, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_jacobian_random(self):
    """Test the Jacobian of the from_euler function."""
    x_init = test_helpers.generate_random_test_euler_angles(dimensions=1)

    self.assert_jacobian_is_correct_fn(rotation_matrix_2d.from_euler, [x_init])

  def test_from_euler_normalized_preset(self):
    """Tests that an angle maps to correct matrix."""
    euler_angles = test_helpers.generate_preset_test_euler_angles(dimensions=1)

    matrix = rotation_matrix_2d.from_euler(euler_angles)

    self.assertAllEqual(
        rotation_matrix_2d.is_valid(matrix),
        np.ones(euler_angles.shape[0:-1] + (1,), dtype=bool))

  @parameterized.parameters(
      ((td.ANGLE_0,), (td.MAT_2D_ID,)),
      ((td.ANGLE_45,), (td.MAT_2D_45,)),
      ((td.ANGLE_90,), (td.MAT_2D_90,)),
      ((td.ANGLE_180,), (td.MAT_2D_180,)),
  )
  def test_from_euler_preset(self, test_inputs, test_outputs):
    """Tests that an angle maps to correct matrix."""
    self.assert_output_is_correct(rotation_matrix_2d.from_euler, test_inputs,
                                  test_outputs)

  @parameterized.parameters(
      ((1,),),
      ((None, 1),),
  )
  def test_from_euler_with_small_angles_approximation_exception_not_raised(
      self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        rotation_matrix_2d.from_euler_with_small_angles_approximation, shapes)

  @parameterized.parameters(
      ("must have exactly 1 dimensions in axis -1", (None,)),)
  def test_from_euler_with_small_angles_approximation_exception_raised(
      self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(
        rotation_matrix_2d.from_euler_with_small_angles_approximation,
        error_msg, shape)

  def test_from_euler_with_small_angles_approximation_random(self):
    """Tests small_angles approximation by comparing to exact calculation."""
    # Only generate small angles. For a test tolerance of 1e-3, 0.17 was found
    # empirically to be the range where the small angle approximation works.
    random_euler_angles = test_helpers.generate_random_test_euler_angles(
        min_angle=-0.17, max_angle=0.17, dimensions=1)

    exact_matrix = rotation_matrix_2d.from_euler(random_euler_angles)
    approximate_matrix = (
        rotation_matrix_2d.from_euler_with_small_angles_approximation(
            random_euler_angles))

    self.assertAllClose(exact_matrix, approximate_matrix, atol=1e-3)

  @parameterized.parameters(
      ((2, 2),),
      ((None, 2, 2),),
  )
  def test_inverse_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_2d.inverse, shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (2,)),
      ("must have exactly 2 dimensions in axis -1", (2, None)),
      ("must have exactly 2 dimensions in axis -2", (None, 2)),
  )
  def test_inverse_exception_raised(self, error_msg, *shapes):
    """Checks the inputs of the inverse function."""
    self.assert_exception_is_raised(rotation_matrix_2d.inverse, error_msg,
                                    shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_preset(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_preset_test_rotation_matrices_2d()

    self.assert_jacobian_is_correct_fn(rotation_matrix_2d.inverse, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_random(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_random_test_rotation_matrix_2d()

    self.assert_jacobian_is_correct_fn(rotation_matrix_2d.inverse, [x_init])

  def test_inverse_random(self):
    """Checks that inverting rotated points results in no transformation."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles(
        dimensions=1)
    tensor_shape = random_euler_angles.shape[:-1]

    random_matrix = rotation_matrix_2d.from_euler(random_euler_angles)
    random_point = np.random.normal(size=tensor_shape + (2,))
    rotated_random_points = rotation_matrix_2d.rotate(random_point,
                                                      random_matrix)
    predicted_invert_random_matrix = rotation_matrix_2d.inverse(random_matrix)
    predicted_invert_rotated_random_points = rotation_matrix_2d.rotate(
        rotated_random_points, predicted_invert_random_matrix)

    self.assertAllClose(
        random_point, predicted_invert_rotated_random_points, rtol=1e-6)

  @parameterized.parameters(
      ((2, 2),),
      ((None, 2, 2),),
  )
  def test_is_valid_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_2d.inverse, shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (2,)),
      ("must have exactly 2 dimensions in axis -1", (2, None)),
      ("must have exactly 2 dimensions in axis -2", (None, 2)),
  )
  def test_is_valid_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(rotation_matrix_2d.is_valid, error_msg,
                                    shape)

  @parameterized.parameters(
      ((2,), (2, 2)),
      ((None, 2), (None, 2, 2)),
      ((1, 2), (1, 2, 2)),
      ((2, 2), (2, 2, 2)),
      ((2,), (1, 2, 2)),
      ((1, 2), (2, 2)),
  )
  def test_rotate_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_2d.rotate, shapes)

  @parameterized.parameters(
      ("must have exactly 2 dimensions in axis -1", (None,), (2, 2)),
      ("must have a rank greater than 1", (2,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (2,), (2, None)),
      ("must have exactly 2 dimensions in axis -2", (2,), (None, 2)),
  )
  def test_rotate_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(rotation_matrix_2d.rotate, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_preset(self):
    """Test the Jacobian of the rotate function."""
    x_matrix_init = test_helpers.generate_preset_test_rotation_matrices_2d()
    tensor_shape = x_matrix_init.shape[:-2] + (2,)
    x_point_init = np.random.uniform(size=tensor_shape)

    self.assert_jacobian_is_correct_fn(rotation_matrix_2d.rotate,
                                       [x_point_init, x_matrix_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_random(self):
    """Test the Jacobian of the rotate function."""
    x_matrix_init = test_helpers.generate_random_test_rotation_matrix_2d()
    tensor_shape = x_matrix_init.shape[:-2] + (2,)
    x_point_init = np.random.uniform(size=tensor_shape)

    self.assert_jacobian_is_correct_fn(rotation_matrix_2d.rotate,
                                       [x_point_init, x_matrix_init])

  @parameterized.parameters(
      ((td.AXIS_2D_0, td.ANGLE_90), (td.AXIS_2D_0,)),
      ((td.AXIS_2D_X, td.ANGLE_90), (td.AXIS_2D_Y,)),
  )
  def test_rotate_preset(self, test_inputs, test_outputs):
    """Tests that the rotate function correctly rotates points."""

    def func(test_point, test_angle):
      random_matrix = rotation_matrix_2d.from_euler(test_angle)
      return rotation_matrix_2d.rotate(test_point, random_matrix)

    self.assert_output_is_correct(func, test_inputs, test_outputs)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for convolution utility functions."""

from absl.testing import parameterized
import numpy as np
from scipy import linalg
import tensorflow as tf

from tensorflow_graphics.geometry.convolution import utils
from tensorflow_graphics.util import test_case


def _dense_to_sparse(data):
  """Converts a numpy array to a tf.SparseTensor."""
  indices = np.where(data)
  return tf.SparseTensor(
      np.stack(indices, axis=-1), data[indices], dense_shape=data.shape)


class UtilsCheckValidGraphConvolutionInputTests(test_case.TestCase):

  def _create_default_tensors_from_shapes(self, shapes):
    """Creates `data`, `sparse`, `sizes` tensors from shapes list."""
    data = tf.convert_to_tensor(
        value=np.random.uniform(size=shapes[0]).astype(np.float32))
    sparse = _dense_to_sparse(np.ones(shape=shapes[1], dtype=np.float32))
    if shapes[2] is not None:
      sizes = tf.convert_to_tensor(
          value=np.ones(shape=shapes[2], dtype=np.int32))
    else:
      sizes = None
    return data, sparse, sizes

  @parameterized.parameters(
      ("'sizes' must have an integer type.", np.float32, np.float32,
       np.float32),
      ("'data' must have a float type.", np.int32, np.float32, np.int32),
      ("'neighbors' and 'data' must have the same type.", np.float32,
       np.float64, np.int32),
  )
  def test_check_valid_graph_convolution_input_exception_raised_types(
      self, err_msg, data_type, neighbors_type, sizes_type):
    """Check the type errors for invalid input types."""
    data = tf.convert_to_tensor(
        value=np.random.uniform(size=(2, 2, 2)).astype(data_type))
    neighbors = _dense_to_sparse(np.ones(shape=(2, 2, 2), dtype=neighbors_type))
    sizes = tf.convert_to_tensor(value=np.array((2, 2), dtype=sizes_type))

    with self.assertRaisesRegexp(TypeError, err_msg):
      utils.check_valid_graph_convolution_input(data, neighbors, sizes)

  @parameterized.parameters(
      (np.float32, np.float32, np.int32),
      (np.float64, np.float64, np.int32),
      (np.float32, np.float32, np.int64),
      (np.float64, np.float64, np.int64),
  )
  def test_check_valid_graph_convolution_input_exception_not_raised_types(
      self, data_type, neighbors_type, sizes_type):
    """Check that no exceptions are raised for valid input types."""
    data = tf.convert_to_tensor(
        value=np.random.uniform(size=(2, 2, 2)).astype(data_type))
    neighbors = _dense_to_sparse(np.ones(shape=(2, 2, 2), dtype=neighbors_type))
    sizes = tf.convert_to_tensor(value=np.array((2, 2), dtype=sizes_type))

    self.assert_exception_is_not_raised(
        utils.check_valid_graph_convolution_input,
        shapes=[],
        data=data,
        neighbors=neighbors,
        sizes=sizes)

  @parameterized.parameters(
      ((2, 3), (2, 2), None),
      ((1, 2, 3), (1, 2, 2), None),
      ((2, 2, 3), (2, 2, 2), None),
      ((1, 2, 3), (1, 2, 2), (1,)),
      ((2, 2, 3), (2, 2, 2), (2,)),
      ((1, 2, 2, 3), (1, 2, 2, 2), (1, 2)),
      ((2, 2, 2, 3), (2, 2, 2, 2), (2, 2)),
  )
  def test_check_valid_graph_convolution_input_exception_not_raised_shapes(
      self, *shapes):
    """Check that valid input shapes do not trigger any exceptions."""
    data, neighbors, sizes = self._create_default_tensors_from_shapes(shapes)

    self.assert_exception_is_not_raised(
        utils.check_valid_graph_convolution_input,
        shapes=[],
        data=data,
        neighbors=neighbors,
        sizes=sizes)

  @parameterized.parameters(
      ((None, 3), (None, 2), None),
      ((1, None, 3), (1, None, None), None),
      ((None, 2, 3), (None, 2, 2), (1,)),
      ((None, None, 3), (None, None, None), (2,)),
      ((1, None, 2, 3), (1, None, None, None), (1, 2)),
  )
  def test_check_valid_graph_convolution_input_exception_not_raised_dynshapes(
      self, *shapes):
    """Check that valid dynamic input shapes do not trigger any exceptions."""
    dtypes = [tf.float32, tf.float32]
    sparse_tensors = [False, True]

    if shapes[2] is not None:
      dtypes.append(tf.int32)
      sparse_tensors.append(False)
      self.assert_exception_is_not_raised(
          utils.check_valid_graph_convolution_input,
          shapes=shapes,
          dtypes=dtypes,
          sparse_tensors=sparse_tensors)
    else:
      self.assert_exception_is_not_raised(
          utils.check_valid_graph_convolution_input,
          shapes=shapes,
          dtypes=dtypes,
          sparse_tensors=sparse_tensors,
          sizes=None)

  def test_check_valid_graph_convolution_dynamic_input_sparse_exception_raised(
      self):
    """Check that passing dense `neighbors` tensor raises exception."""
    error_msg = "must be a SparseTensor"
    dtypes = [tf.float32, tf.float32, tf.int32]
    sparse_tensors = [False, False, False]
    shapes = ((None, 3), (None, 2), (None,))

    self.assert_exception_is_raised(
        utils.check_valid_graph_convolution_input,
        error_msg,
        shapes=shapes,
        dtypes=dtypes,
        sparse_tensors=sparse_tensors)

  @parameterized.parameters(
      ("must have a rank of 2", (5, 2), (1, 5, 2), None),
      ("must have a rank greater than 1", (5,), (5, 5), None),
      ("must have a rank of 2", (5, 2), (5,), None),
      ("must have a rank of 3", (5, 5, 2), (5, 5), None),
      ("must have the same number of dimensions in axes", (3, 2), (3, 2), None),
      ("must have a rank of 1", (5, 5, 2), (5, 5, 5), (5, 5)),
      ("Not all batch dimensions are identical.", (1, 5, 2), (1, 5, 5), (2,)),
  )
  def test_check_valid_graph_convolution_input_exception_raised_shapes(
      self, error_msg, *shapes):
    """Check that invalid input shapes trigger the right exceptions."""
    data, neighbors, sizes = self._create_default_tensors_from_shapes(shapes)

    self.assert_exception_is_raised(
        utils.check_valid_graph_convolution_input,
        error_msg,
        shapes=[],
        data=data,
        neighbors=neighbors,
        sizes=sizes)


class UtilsCheckValidGraphPoolingInputTests(test_case.TestCase):

  @parameterized.parameters(
      ("'sizes' must have an integer type.", np.float32, np.float32,
       np.float32),
      ("'data' must have a float type.", np.int32, np.float32, np.int32),
      ("'pool_map' and 'data' must have the same type.", np.float32, np.float64,
       np.int32),
  )
  def test_check_valid_graph_pooling_exception_raised_types(
      self, err_msg, data_type, pool_map_type, sizes_type):
    """Check the type errors for invalid input types."""
    data = tf.convert_to_tensor(value=np.ones((2, 3, 3), dtype=data_type))
    pool_map = _dense_to_sparse(np.ones((2, 3, 3), dtype=pool_map_type))
    sizes = tf.convert_to_tensor(
        value=np.array(((1, 2), (2, 3)), dtype=sizes_type))

    with self.assertRaisesRegexp(TypeError, err_msg):
      utils.check_valid_graph_pooling_input(data, pool_map, sizes)

  @parameterized.parameters(
      (np.float32, np.float32, np.int32),
      (np.float64, np.float64, np.int32),
      (np.float32, np.float32, np.int64),
      (np.float64, np.float64, np.int64),
  )
  def test_check_valid_graph_pooling_exception_not_raised_types(
      self, data_type, pool_map_type, sizes_type):
    """Check there are no exceptions for valid input types."""
    data = tf.convert_to_tensor(value=np.ones((2, 3, 3), dtype=data_type))
    pool_map = _dense_to_sparse(np.ones((2, 3, 3), dtype=pool_map_type))
    sizes = tf.convert_to_tensor(
        value=np.array(((1, 2), (2, 3)), dtype=sizes_type))

    self.assert_exception_is_not_raised(
        utils.check_valid_graph_pooling_input,
        shapes=[],
        data=data,
        pool_map=pool_map,
        sizes=sizes)

  @parameterized.parameters(
      ((2, 3), (4, 2), None),
      ((1, 2, 3), (1, 5, 2), None),
      ((2, 2, 3), (2, 5, 2), ((3, 2), (2, 5))),
  )
  def test_check_valid_graph_pooling_exception_not_raised_shapes(
      self, data_shape, pool_map_shape, sizes):
    """Check that valid input shapes do not trigger any exceptions."""
    data = tf.convert_to_tensor(value=np.ones(data_shape, dtype=np.float32))
    pool_map = _dense_to_sparse(np.ones(pool_map_shape, dtype=np.float32))
    sizes = sizes if sizes is None else tf.convert_to_tensor(value=sizes)

    self.assert_exception_is_not_raised(
        utils.check_valid_graph_pooling_input,
        shapes=[],
        data=data,
        pool_map=pool_map,
        sizes=sizes)

  @parameterized.parameters(
      ((None, 3), (None, 2), None),
      ((1, None, 3), (1, None, None), None),
      ((None, 2, 3), (None, 5, 2), (2, 2)),
  )
  def test_check_valid_graph_pooling_exception_not_raised_dynamic_shapes(
      self, *shapes):
    """Check that valid dynamic input shapes do not trigger any exceptions."""
    dtypes = [tf.float32, tf.float32]
    sparse_tensors = [False, True]

    if shapes[2] is not None:
      dtypes.append(tf.int32)
      sparse_tensors.append(False)
      self.assert_exception_is_not_raised(
          utils.check_valid_graph_pooling_input,
          shapes=shapes,
          dtypes=dtypes,
          sparse_tensors=sparse_tensors)
    else:
      self.assert_exception_is_not_raised(
          utils.check_valid_graph_pooling_input,
          shapes=shapes,
          dtypes=dtypes,
          sparse_tensors=sparse_tensors,
          sizes=None)

  def test_check_graph_pooling_input_sparse_exception_raised(self):
    """Check that passing dense `neighbors` tensor raises exception."""
    error_msg = "must be a SparseTensor"
    dtypes = [tf.float32, tf.float32, tf.int32]
    sparse_tensors = [False, False, False]
    shapes = ((2, 2, 3), (2, 5, 2), (2, 2))
    self.assert_exception_is_raised(
        utils.check_valid_graph_convolution_input,
        error_msg,
        shapes=shapes,
        dtypes=dtypes,
        sparse_tensors=sparse_tensors)

  @parameterized.parameters(
      ("must have a rank greater than 1", (5,), (5, 5), None),
      ("must have a rank of 2", (5, 2), (5,), None),
      ("must have the same number of dimensions in axes", (3, 2), (3, 2), None),
      ("Not all batch dimensions are identical.", (3, 5, 2), (1, 5, 5), None),
      ("must have a rank of 2", (2, 5, 2), (2, 3, 5), (3, 5)),
      ("Not all batch dimensions are identical.", (3, 5, 2), (3, 3, 5),
       ((3, 5), (2, 4))),
  )
  def test_check_valid_graph_pooling_exception_raised_shapes(
      self, err_msg, data_shape, pool_map_shape, sizes):
    """Check that invalid input shapes trigger the right exceptions."""
    data = tf.convert_to_tensor(value=np.ones(data_shape, dtype=np.float32))
    pool_map = _dense_to_sparse(np.ones(pool_map_shape, dtype=np.float32))
    sizes = sizes if sizes is None else tf.convert_to_tensor(value=sizes)

    self.assert_exception_is_raised(
        utils.check_valid_graph_pooling_input,
        err_msg,
        shapes=[],
        data=data,
        pool_map=pool_map,
        sizes=sizes)


class UtilsFlattenBatchTo2dTests(test_case.TestCase):

  @parameterized.parameters(((5, 3),), ((3,),))
  def test_input_rank_exception_raised(self, *shapes):
    """Check that invalid input data rank triggers the right exceptions."""
    self.assert_exception_is_raised(utils.flatten_batch_to_2d,
                                    "must have a rank greater than 2", shapes)

  def test_flatten_batch_to_2d_exception_raised_types(self):
    """Check the exception when input is not an integer."""
    with self.assertRaisesRegexp(TypeError,
                                 "'sizes' must have an integer type."):
      utils.flatten_batch_to_2d(np.ones((3, 4, 3)), np.ones((3,)))

  @parameterized.parameters(
      ((None, 3, 3), None),
      ((3, None, 3), (3,)),
  )
  def test_check_flatten_batch_to_2d_exception_not_raised_dynamic_shapes(
      self, *shapes):
    """Check that valid dynamic input shapes do not trigger any exceptions."""
    dtypes = [tf.float32]

    if shapes[1] is not None:
      dtypes.append(tf.int32)
      self.assert_exception_is_not_raised(
          utils.flatten_batch_to_2d, shapes=shapes, dtypes=dtypes)
    else:
      self.assert_exception_is_not_raised(
          utils.flatten_batch_to_2d, shapes=shapes, dtypes=dtypes, sizes=None)

  @parameterized.parameters(
      ("must have a rank of 1", (3, 4, 3), (3, 4)),
      ("must have a rank of 1", (3, 4, 5), (3, 4, 5)),
  )
  def test_flatten_batch_to_2d_exception_raised(self, error_msg, *shapes):
    """Check the exception when the shape of 'sizes' is invalid."""
    self.assert_exception_is_raised(
        utils.flatten_batch_to_2d,
        error_msg,
        shapes,
        dtypes=(tf.float32, tf.int32))

  def test_flatten_batch_to_2d_random(self):
    """Test flattening with random inputs."""
    ndims_batch = np.random.randint(low=1, high=5)
    batch_dims = np.random.randint(low=1, high=10, size=ndims_batch)
    data_dims = np.random.randint(low=1, high=20, size=2)
    dims = np.concatenate([batch_dims, data_dims], axis=0)
    data = np.random.uniform(size=dims)

    with self.subTest(name="random_padding"):
      sizes = np.random.randint(low=0, high=data_dims[0], size=batch_dims)
      y, unflatten = utils.flatten_batch_to_2d(data, sizes)
      data_unflattened = unflatten(y)

      self.assertAllEqual(tf.shape(input=y), [np.sum(sizes), data_dims[1]])
      self.assertAllEqual(
          tf.shape(input=data_unflattened), tf.shape(input=data))

    with self.subTest(name="no_padding_with_sizes"):
      sizes = data_dims[0] * np.ones_like(sizes, dtype=np.int32)
      y, unflatten = utils.flatten_batch_to_2d(data, sizes)

      self.assertAllEqual(tf.shape(input=y), [np.sum(sizes), data_dims[1]])
      self.assertAllEqual(data, unflatten(y))

    with self.subTest(name="no_padding_with_sizes_none"):
      y, unflatten = utils.flatten_batch_to_2d(data, sizes=None)

      self.assertAllEqual(tf.shape(input=y), [np.sum(sizes), data_dims[1]])
      self.assertAllEqual(data, unflatten(y))

  def test_flatten_batch_to_2d_zero_sizes(self):
    """Test flattening with zero sizes."""
    data = np.ones(shape=(10, 5, 3, 2))
    sizes = np.zeros(shape=(10, 5), dtype=np.int32)

    y, unflatten = utils.flatten_batch_to_2d(data, sizes)

    self.assertAllEqual([0, 2], tf.shape(input=y))
    self.assertAllEqual(np.zeros_like(data), unflatten(y))

  def test_flatten_batch_to_2d_unflatten_different_feature_dims(self):
    """Test when inputs to flattening/unflattening use different channels."""
    data_in = np.random.uniform(size=(3, 1, 7, 5, 4))
    data_out = np.concatenate([data_in, data_in], axis=-1)

    y, unflatten = utils.flatten_batch_to_2d(data_in)

    self.assertAllEqual(unflatten(tf.concat([y, y], axis=-1)), data_out)

  def test_flatten_batch_to_2d_jacobian_random(self):
    """Test the jacobian is correct for random inputs."""
    data_init = np.random.uniform(size=(3, 2, 7, 5, 4))
    sizes = np.random.randint(low=1, high=5, size=(3, 2, 7))
    flat_init = np.random.uniform(size=(np.sum(sizes), 10))

    def flatten_batch_to_2d(data):
      flattened, _ = utils.flatten_batch_to_2d(data, sizes=sizes)
      return flattened

    def unflatten_2d_to_batch(flat):
      _, unflatten = utils.flatten_batch_to_2d(data_init, sizes=sizes)
      return unflatten(flat)

    with self.subTest(name="flatten"):
      self.assert_jacobian_is_correct_fn(flatten_batch_to_2d, [data_init])

    with self.subTest(name="unflatten"):
      self.assert_jacobian_is_correct_fn(unflatten_2d_to_batch, [flat_init])

  @parameterized.parameters((np.int32), (np.float32), (np.uint16))
  def test_flatten_batch_to_2d_unflatten_types(self, dtype):
    """Test unflattening with int and float types."""
    data = np.ones(shape=(2, 2, 3, 2), dtype=dtype)
    sizes = ((3, 2), (1, 3))
    desired_unflattened = data
    desired_unflattened[0, 1, 2, :] = 0
    desired_unflattened[1, 0, 1:, :] = 0

    flat, unflatten = utils.flatten_batch_to_2d(data, sizes=sizes)
    data_unflattened = unflatten(flat)

    self.assertEqual(data.dtype, data_unflattened.dtype.as_numpy_dtype)
    self.assertAllEqual(data_unflattened, desired_unflattened)


class UtilsUnflatten2dToBatchTest(test_case.TestCase):

  @parameterized.parameters(((3, 2, 4), (3,)), ((5,), (4, 2)))
  def test_input_rank_exception_raised(self, *shapes):
    """Check that invalid inputs trigger the right exception."""
    self.assert_exception_is_raised(utils.unflatten_2d_to_batch,
                                    "data must have a rank of 2", shapes)

  def test_input_type_exception_raised(self):
    """Check that invalid input types trigger the right exception."""
    with self.assertRaisesRegexp(TypeError,
                                 "'sizes' must have an integer type."):
      utils.unflatten_2d_to_batch(np.ones((3, 4)), np.ones((3,)))

  @parameterized.parameters(
      ((3, 2, 1), None, 5),
      ((3, 2, 1, 2), 4, 2),
      (((3, 2), (1, 2)), None, 2),
  )
  def test_unflatten_batch_to_2d_random(self, sizes, max_rows, num_features):
    """Test unflattening with random inputs."""
    max_rows = np.max(sizes) if max_rows is None else max_rows
    output_shape = np.concatenate(
        (np.shape(sizes), (max_rows,), (num_features,)))
    total_rows = np.sum(sizes)
    data = 0.1 + np.random.uniform(size=(total_rows, num_features))

    unflattened = utils.unflatten_2d_to_batch(data, sizes, max_rows)
    flattened = tf.reshape(unflattened, (-1, num_features))
    nonzero_rows = tf.compat.v1.where(tf.norm(tensor=flattened, axis=-1))
    flattened_unpadded = tf.gather(
        params=flattened, indices=tf.squeeze(input=nonzero_rows, axis=-1))

    self.assertAllEqual(tf.shape(input=unflattened), output_shape)
    self.assertAllEqual(flattened_unpadded, data)

  def test_unflatten_batch_to_2d_preset(self):
    """Test unflattening with a preset input."""
    data = 1. + np.reshape(np.arange(12, dtype=np.float32), (6, 2))
    sizes = (2, 3, 1)
    output_true = np.array(
        (((1., 2.), (3., 4.), (0., 0.)), ((5., 6.), (7., 8.), (9., 10.)),
         ((11., 12.), (0., 0.), (0., 0.))),
        dtype=np.float32)
    output_true_padded = np.pad(
        output_true, ((0, 0), (0, 2), (0, 0)), mode="constant")

    output = utils.unflatten_2d_to_batch(data, sizes, max_rows=None)
    output_padded = utils.unflatten_2d_to_batch(data, sizes, max_rows=5)

    self.assertAllEqual(output, output_true)
    self.assertAllEqual(output_padded, output_true_padded)

  @parameterized.parameters(
      ((3, 2, 1), None, 5),
      ((3, 2, 1, 2), 4, 2),
      (((3, 2), (1, 2)), None, 2),
  )
  def test_unflatten_batch_to_2d_jacobian_random(self, sizes, max_rows,
                                                 num_features):
    """Test that the jacobian is correct."""
    max_rows = np.max(sizes) if max_rows is None else max_rows
    total_rows = np.sum(sizes)
    data_init = 0.1 + np.random.uniform(size=(total_rows, num_features))

    def unflatten_2d_to_batch(data):
      return utils.unflatten_2d_to_batch(data, sizes, max_rows)

    self.assert_jacobian_is_correct_fn(unflatten_2d_to_batch, [data_init])

  @parameterized.parameters((np.int32), (np.float32), (np.uint16))
  def test_unflatten_batch_to_2d_types(self, dtype):
    """Test unflattening with int and float types."""
    data = np.ones(shape=(6, 2), dtype=dtype)
    sizes = (2, 2, 2)
    unflattened_true = np.ones(shape=(3, 2, 2), dtype=dtype)

    unflattened = utils.unflatten_2d_to_batch(data, sizes)

    self.assertEqual(data.dtype, unflattened.dtype.as_numpy_dtype)
    self.assertAllEqual(unflattened, unflattened_true)


class UtilsConvertToBlockDiag2dTests(test_case.TestCase):

  def _validate_sizes(self, block_diag_tensor, sizes):
    """Assert all elements outside the blocks are zero."""
    data = [np.ones(shape=s) for s in sizes]
    mask = 1.0 - linalg.block_diag(*data)

    self.assertAllEqual(
        tf.sparse.to_dense(block_diag_tensor) * mask, np.zeros_like(mask))

  def test_convert_to_block_diag_2d_exception_raised_types(self):
    """Check the exception when input is not a SparseTensor."""
    with self.assertRaisesRegexp(TypeError, "'data' must be a 'SparseTensor'."):
      utils.convert_to_block_diag_2d(np.zeros(shape=(3, 3, 3)))

    with self.assertRaisesRegexp(TypeError,
                                 "'sizes' must have an integer type."):
      utils.convert_to_block_diag_2d(
          _dense_to_sparse(np.ones(shape=(3, 3, 3))),
          np.ones(shape=(3, 2)),
      )

  def test_convert_to_block_diag_2d_exception_raised_ranks(self):
    """Check the exception when input data rank is invalid."""
    with self.assertRaisesRegexp(ValueError, "must have a rank greater than 2"):
      utils.convert_to_block_diag_2d(_dense_to_sparse(np.ones(shape=(3, 3))))

    with self.assertRaisesRegexp(ValueError, "must have a rank greater than 2"):
      utils.convert_to_block_diag_2d(_dense_to_sparse(np.ones(shape=(3,))))

  def test_convert_to_block_diag_2d_exception_raised_sizes(self):
    """Check the expetion when the shape of sizes is invalid."""
    with self.assertRaisesRegexp(ValueError, "must have a rank of 2"):
      utils.convert_to_block_diag_2d(
          _dense_to_sparse(np.ones(shape=(3, 3, 3))),
          np.ones(shape=(3,), dtype=np.int32))

    with self.assertRaisesRegexp(ValueError, "must have a rank of 3"):
      utils.convert_to_block_diag_2d(
          _dense_to_sparse(np.ones(shape=(4, 3, 3, 3))),
          np.ones(shape=(4, 3), dtype=np.int32))

    with self.assertRaisesRegexp(ValueError,
                                 "must have exactly 2 dimensions in axis -1"):
      utils.convert_to_block_diag_2d(
          _dense_to_sparse(np.ones(shape=(3, 3, 3))),
          np.ones(shape=(3, 1), dtype=np.int32))

  def test_convert_to_block_diag_2d_random(self):
    """Test block diagonalization with random inputs."""
    sizes = np.random.randint(low=2, high=6, size=(3, 2))
    data = [np.random.uniform(size=s) for s in sizes]
    batch_data_padded = np.zeros(
        shape=np.concatenate(([len(sizes)], np.max(sizes, axis=0)), axis=0))
    for i, s in enumerate(sizes):
      batch_data_padded[i, :s[0], :s[1]] = data[i]

    batch_data_padded_sparse = _dense_to_sparse(batch_data_padded)
    block_diag_data = linalg.block_diag(*data)
    block_diag_sparse = utils.convert_to_block_diag_2d(
        batch_data_padded_sparse, sizes=sizes)

    self.assertAllEqual(tf.sparse.to_dense(block_diag_sparse), block_diag_data)

  def test_convert_to_block_diag_2d_no_padding(self):
    """Test block diagonalization without any padding."""
    batch_data = np.random.uniform(size=(3, 4, 5, 4))
    block_diag_data = linalg.block_diag(
        *[x for x in np.reshape(batch_data, (-1, 5, 4))])

    batch_data_sparse = _dense_to_sparse(batch_data)
    block_diag_sparse = utils.convert_to_block_diag_2d(batch_data_sparse)

    self.assertAllEqual(tf.sparse.to_dense(block_diag_sparse), block_diag_data)

  def test_convert_to_block_diag_2d_validate_indices(self):
    """Test block diagonalization when we filter out out-of-bounds indices."""
    sizes = ((2, 3), (2, 3), (2, 3))

    batch = _dense_to_sparse(np.random.uniform(size=(3, 4, 3)))
    block_diag = utils.convert_to_block_diag_2d(batch, sizes, True)

    self._validate_sizes(block_diag, sizes)

  def test_convert_to_block_diag_2d_large_sizes(self):
    """Test when the desired blocks are larger than the data shapes."""
    sizes = ((5, 5), (6, 6), (7, 7))

    batch = _dense_to_sparse(np.random.uniform(size=(3, 4, 3)))
    block_diag = utils.convert_to_block_diag_2d(batch, sizes)

    self._validate_sizes(block_diag, sizes)

  def test_convert_to_block_diag_2d_batch_shapes(self):
    """Test with different batch shapes."""
    sizes_one_batch_dim = np.concatenate(
        [np.random.randint(low=1, high=h, size=(6 * 3 * 4, 1)) for h in (5, 7)],
        axis=-1)
    data = [np.random.uniform(size=s) for s in sizes_one_batch_dim]
    data_one_batch_dim_padded = np.zeros(shape=(6 * 3 * 4, 5, 7))
    for i, s in enumerate(sizes_one_batch_dim):
      data_one_batch_dim_padded[i, :s[0], :s[1]] = data[i]
    data_many_batch_dim_padded = np.reshape(data_one_batch_dim_padded,
                                            (6, 3, 4, 5, 7))
    sizes_many_batch_dim = np.reshape(sizes_one_batch_dim, (6, 3, 4, -1))

    data_one_sparse = _dense_to_sparse(data_one_batch_dim_padded)
    data_many_sparse = _dense_to_sparse(data_many_batch_dim_padded)
    one_batch_dim = utils.convert_to_block_diag_2d(data_one_sparse,
                                                   sizes_one_batch_dim)
    many_batch_dim = utils.convert_to_block_diag_2d(data_many_sparse,
                                                    sizes_many_batch_dim)

    self.assertAllEqual(
        tf.sparse.to_dense(one_batch_dim), tf.sparse.to_dense(many_batch_dim))
    self._validate_sizes(one_batch_dim, sizes_one_batch_dim)

  def test_convert_to_block_diag_2d_jacobian_random(self):
    """Test the jacobian is correct with random inputs."""
    sizes = np.random.randint(low=2, high=6, size=(3, 2))
    data = [np.random.uniform(size=s) for s in sizes]
    batch_data_padded = np.zeros(
        shape=np.concatenate([[len(sizes)], np.max(sizes, axis=0)], axis=0))
    for i, s in enumerate(sizes):
      batch_data_padded[i, :s[0], :s[1]] = data[i]
    sparse_ind = np.where(batch_data_padded)
    sparse_val_init = batch_data_padded[sparse_ind]

    def convert_to_block_diag_2d(sparse_val):
      sparse = tf.SparseTensor(
          np.stack(sparse_ind, axis=-1), sparse_val, batch_data_padded.shape)
      return utils.convert_to_block_diag_2d(sparse, sizes).values

    self.assert_jacobian_is_correct_fn(convert_to_block_diag_2d,
                                       [sparse_val_init])


if __name__ == "__main__":
  test_case.main()
